package dans.algorithm;

import dans.network.BackEndMessage;
import dans.network.MessageInQueue;
import dans.network.NetworkObject;
import dans.network.NetworkState;
import dans.network.Processor;
import dans.util.Log;
import dans.util.Utils;
import java.util.Random;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This abstract class should be extended to create new algorithms to run in DANS.
 * 
 * A {@link #algorithm()} method must be provided that returns the results of the
 * algorithm as a Java Object. This method should contain a main loop that runs
 * until {@link #doMainLoop()} returns false. For example:
 * 
 * <pre>
 * {@code
 * public Object algorithm() {
 *  //Setup anything needed for the algorithm
 * 
 *  while(doMainLoop) {
 *      //Do some calls to send and receive here.
 *  }
 * 
 *  //Running anything before the algorthim terminates
 * 
 *  return someResultHere
 * }
 * }
 * </pre>
 * Implementing classes should only reference or use classes or methods from the 
 * dans.algorithm and dans.utils packages. Use of any other class from the dans
 * library can break the distributed algorithm model and/or the simulator.
 * 
 * No methods in this class should be called in the constructor of implementing
 * child classes. Doing so will result in a null value or equivalent being 
 * returned and possible exceptions. This may also break the simulation.
 * 
 * @author Daniel Servos
 * @version 0.1beta
 */
public abstract class Algorithm extends Thread {
    private final static int THREAD_MIN_WAIT_TIME = 100;
    private final static int THREAD_RETRY_TIME = 300;
    
    /**
     * No failure has occurred.
     */
    public final static int FAIL_NONE = 0;
    /**
     * A clean failure has occurred.
     */
    public final static int FAIL_CLEAN = 1;
    /**
     * A Byzantine failure has occurred.
     */
    public final static int FAIL_BYZAN = 2;
    
    private Processor p = null;
    private NetworkState ns = null;
    private ConcurrentHashMap<String, Object> memory = null;
    private int port = -1;
    private boolean terminated = false;
    private String status = NetworkObject.STATUS_NOTSTARTED;
    private Object commonLock = null;
    private final Object statusLock = new Object();
    private boolean paused = false;
    private boolean firstloop = true;
    private Random rand;
    
    private int failType = FAIL_NONE;
    private final Object failLock = new Object();
    
    private final void setUp(Processor p, NetworkState ns, Integer port, ConcurrentHashMap<String, Object> memory){
        this.p = p;
        this.ns = ns;
        this.port = port;
        this.memory = memory;
        commonLock = ns.getCommonLock();
        //TODO: Let user pick seed
        rand = new Random();
    }
    
    /**
     * Gets a reference to a ConcurrentHashMap that represents the shared memory 
     * of the processor the algorithm is running on. This shared memory object 
     * allows algorithms on the same processor to communicate and share resources.
     * While the HashMap its self is thread safe it is up to the algorithm to ensure
     * any objects or values stored in the HashMap are properly synchronized.
     * 
     * @return a ConcurrentHashMap instance that is shared with all other algorithms running on the same processor.
     */
    public ConcurrentHashMap<String, Object> getMemory() {
        return memory;
    }
    
    /**
     * Gets a value indicating how this algorithm failed.
     * 
     * @return one of {@link #FAIL_NONE}, {@link #FAIL_CLEAN}, or {@link #FAIL_BYZAN}.
     */
    public int getFailType() {
        synchronized(failLock) {
            return failType;
        }
    }
    
    /**
     * Gets the current tick the simulation is on.
     * 
     * @return the current tick value.
     */
    public final int getTick() {
        return ns.getTick();
    }
    
    /**
     * Gets the default port number for the simulation. This is not necessarily
     * the port the algorithm is listening on. Use {@link #getPort()} to find the
     * port the algorithm is listening on.
     * 
     * @return a port number.
     * @see #getPort() 
     */
    public final int getDefaultPort() {
        return (int) ns.getNetworkSetting("defaultPort");
    }
    
    /**
     * Gets the port the algorithm is listening on. The algorithm will only
     * receive messages sent to the processor on this port.
     * 
     * @return a port number.
     */
    public final int getPort() {
        return port;
    }
    
    /**
     * Gets the bandwidth of a given link. Returns -1 if there is no link from 
     * this processor to another with the given ID.
     * 
     * @param linkID a link ID.
     * @return the bandwidth of the link in messages per tick or -1.
     */
    public final double getLinkBandwidth(String linkID) {
        return p.getLinkBandwidth(ns, linkID);
    }
    
    /**
     * Gets the delay of the given link. Returns -1 if there is no link from this
     * processor to another with the given ID.
     * 
     * @param linkID a link ID.
     * @return the delay of the link in ticks or -1.
     */
    public final int getLinkDelay(String linkID) {
        return p.getLinkDelay(ns, linkID);
    }
    
    /**
     * Checks if the processor has a neighbor with the given ID. A neighbor is
     * considered to be a processor that the current processor has a directional
     * link to. Processors that have a directional link to this processor but this
     * processor does not have a directional link back to are NOT counted as neighbors.
     * 
     * @param neighborID the processor ID to check.
     * @return true if the processor has a directional link to the given processor. Otherwise false.
     * @see #hasLink(java.lang.String) 
     */
    public final boolean hasNeighbor(String neighborID) {
        return p.hasNeighbour(neighborID);
    }
    
    /**
     * Checks if the processor has a link with the given ID. Only links leaving
     * from this processor to another are considered.
     * 
     * @param linkID the link ID to check.
     * @return true if the processor has a link with the given ID. Otherwise false.
     * @see #hasNeighbor(java.lang.String) 
     */
    public final boolean hasLink(String linkID) {
        return p.hasLink(linkID);
    }
    
    /**
     * Finds the neighbor that a given link points to.
     * 
     * @param linkID the ID of the link to find a processor down.
     * @return a processor ID for the processor this link points to or null if the link could not be found.
     * @see #getNeighborsLink(java.lang.String) 
     */
    public final String getNeighborDownLink(String linkID) {
        return p.getNeighbourID(linkID);
    }
    
    /**
     * Finds a link pointing to the given neighbor.
     * 
     * @param neighborID the ID of the neighbor to find a link to.
     * @return a link ID or null if no such link could be found.
     * @see #getNeighborDownLink(java.lang.String) 
     */
    public final String getNeighborsLink(String neighborID) {
        return p.getLinkID(neighborID);
    }
    
    /**
     * Gets the set of neighbors for this processor. A neighbor is
     * considered to be a processor that the current processor has a directional
     * link to. Processors that have a directional link to this processor but this
     * processor does not have a directional link back to are NOT counted as neighbors.
     * 
     * @return an array of processor IDs.
     * @see #getLinks() 
     */
    public final String[] getNeighbors() {
        return p.getNeighbourIDs();
    }
    
    /**
     * Gets the set of directional links leaving this processor. This does not
     * include links that only point to this processor.
     * 
     * @return an array of link IDs.
     * @see #getNeighbors() 
     */
    public final String[] getLinks() {
        return p.getLinkIDs();
    }
    
    /**
     * Gets the set of processors that have a directional link to this processor.
     * This is not necessarily the same as set returned by {@link #getNeighbors()}
     * as it only includes processors that have a link TO this processor and not
     * necessarily FROM this processor. If the network is setup in a directional 
     * tree such that all child processor point to their parent, this set would
     * contain the children of the processor.
     * 
     * @return an array of processor IDs.
     * @see #getNeighbors() 
     * @see #getSourceLinks() 
     */
    public final String[] getSources() {
        return p.getSourceIDs();
    }
    
    /**
     * Gets the set of links that point to this processor. This set does not
     * include any links FROM this processor TO other processors, only directional
     * links that point TO this processor.
     * 
     * @return an array of link IDs.
     * @see #getSources()
     * @see #getLinks() 
     */
    public final String[] getSourceLinks() {
        return p.getSourceLinkIDs();
    }
    
    /**
     * Gets the ID of the processor the algorithm is running on. This ID is not
     * guaranteed to be an Integer or even a number. Note that the function getId() 
     * (with a lower case 'd') is not the same as this function and should not be 
     * used.
     * 
     * @return the ID of the processor.
     */
    public final String getID() {
        return p.getID();
    }
    
    /**
     * Checks if this processor is set as the root of the network graph. Currently
     * this feature is not supported in the GUI front-end so this function should
     * not be used. Instead the ID of the root processor should be set to something
     * like "root" and {@link #getID()} should be used to check if the processor
     * has this ID.
     * 
     * @return true if the processor is set as the root. Otherwise false.
     * @see #getID() 
     * @deprecated this method does not currently work correctly with the front-end.
     */
    public final boolean isRoot() {
        return p.isRoot();
    }
    
    /**
     * This method should be overridden by any implementing child class. It is
     * called when the processor is started and should contain a loop that runs
     * until {@link #doMainLoop()} returns false. If the algorithm is finished it
     * can break or return out of this loop immediately with out checking 
     * {@link #doMainLoop()}. All calls to {@link #send(dans.algorithm.Message)} 
     * and {@link #receive()} should be contained in such a loop. A value maybe
     * returned and will be displayed next to the node in the front-end when the
     * algorithm terminates. Otherwise null can be returned. See the description
     * of the {@link Algorithm} class for more details.
     * 
     * This method should not be called directly, just overridden by a child class.
     * 
     * @return any value with a toString method or null.
     * @see Algorithm 
     */
    public abstract Object algorithm();
    
    /**
     * This is an internal method that should not be called directly.
     */
    @Override
    public final void run() {
        Log.debug("Algorithm started on port " + getPort(), "P:" + getID() + "#" + getPort());
        synchronized(statusLock) {
            status = NetworkObject.STATUS_RUNNING;
            p.algorithmUpdateStatus(this, status);
        }
        
        Object returned = null;
        
        try {
            returned = algorithm();
        } catch (Exception e) {
            //TODO: Handel algorithm expections.
            Log.error("Algorithm exception.", e, "P:" + getID() + "#" + getPort());
            synchronized(statusLock) {
                status = NetworkObject.STATUS_ERROR;
                p.algorithmUpdateStatus(this, status);
            }
        }

        synchronized(statusLock) {
            if(status.equals(NetworkObject.STATUS_RUNNING) || status.equals(NetworkObject.STATUS_WAITING) || status.equals(NetworkObject.STATUS_PAUSED)) {
                status = NetworkObject.STATUS_TERMINATED;
                p.algorithmUpdateStatus(this, status);
            }
        }
        
        terminated = true;
        p.algorithmTerminated(this);
        Log.debug("Algorithm terminated on port " + getPort(), "P:" + getID() + "#" + getPort());
        
        if(returned != null) {
            display(returned.toString(), 1);
        }
    }
    
    /**
     * Gets the current status of the algorithm. This method is not intended to 
     * be called from a child class.
     * 
     * @return a string containing the status of the algorithm.
     */
    public final String getStatus() {
        synchronized(statusLock) {
            return status;
        }
    }
    
    /**
     * Writes some text to the log. This is a thread safe way for the algorithm
     * to output text to the console and logging system. This method should be
     * used in place of System.out.print or System.out.println.
     * 
     * @param msg the text to be written to the log.
     * @see #display(java.lang.String) 
     */
    public final void print(String msg) {
        Log.info(msg, "P:" + getID() + "#" + getPort());
    }
    
    /**
     * Displays some text above the processor in the GUI front-end. The text should
     * be short.
     * 
     * @param text the text to display.
     * @see #print(java.lang.String) 
     * @see #display(java.lang.String, int) 
     */
    public final void display(String text) {
        display(text, 0);
    }
    
    /**
     * Displays some text above or below the processor in the GUI front-end. If a
     * pos of 0 is given the text will be displayed above the processor, if a pos
     * of 1 is given the text will be displayed under the processor. Note that the
     * result of the algorithm will overwrite any text displayed under the processor
     * when the algorithm terminates. The text should be short.
     * 
     * @param text the text to display.
     * @param pos the position to display the text in. 0 for above, 1 for below, any other value is equivalent to calling {@link #print(java.lang.String)}.
     * @see #print(java.lang.String)
     * @see #display(java.lang.String)
     */
    public final void display(String text, int pos) {
        p.displayEvent(text, pos, this);
    }
    
    /**
     * Sends the given message. This should be a new and unsent {@link Message}
     * object. The {@link Message} object must contain either a processor ID or
     * link ID to send the message to.
     * 
     * This method does NOT block or wait for the message to be delivered. Depending
     * on the failure settings delivery is not necessarily guaranteed. This method
     * returning true does not necessarily guarantee the message was delivered.
     * 
     * @param msg a new {@link Message} object that has not been sent.
     * @return true if the message was successfully added to the links output queue. Otherwise false if the send operation fails (e.g. if there is no link from this processor to one with the given processor ID).
     * @see Message
     * @see #send(java.lang.String, java.lang.String) 
     * @see #send(java.lang.String, java.lang.String, int) 
     * @see #sendLink(java.lang.String, java.lang.String) 
     * @see #sendLink(java.lang.String, java.lang.String, int) 
     */
    public final boolean send(Message msg) {
        BackEndMessage bm = new BackEndMessage(msg);
        bm.fromID = getID();
        bm.senderPort = getPort();
        if(bm.port < 0) {
            bm.port = getPort();
        }
        
        return p.send(ns, bm, this);
    }
    
    /**
     * Sends the given message to a processor with the given ID.
     * 
     * This method does NOT block or wait for the message to be delivered. Depending
     * on the failure settings delivery is not necessarily guaranteed. This method
     * returning true does not necessarily guarantee the message was delivered.
     * 
     * @param id the ID of the processor to send the message to.
     * @param msg the message to send to the processor.
     * @return true if the message was successfully added to the links output queue. Otherwise false if the send operation fails (e.g. if there is no link from this processor to one with the given processor ID).
     * @see #send(dans.algorithm.Message) 
     * @see #send(java.lang.String, java.lang.String, int) 
     * @see #sendLink(java.lang.String, java.lang.String) 
     * @see #sendLink(java.lang.String, java.lang.String, int) 
     */
    public final boolean send(String id, String msg) {
        return send(id, msg, -1);
    }
    
    /**
     * Sends the given message to a processor with the given ID on the given port.
     * Specifying a port allows this algorithm to send messages to algorithms running on
     * different ports on different processors.
     * 
     * This method does NOT block or wait for the message to be delivered. Depending
     * on the failure settings delivery is not necessarily guaranteed. This method
     * returning true does not necessarily guarantee the message was delivered.
     * 
     * @param id the ID of the processor to send the message to.
     * @param msg the message to send to the processor.
     * @param port the port to send the message to on the receiving processor.
     * @return true if the message was successfully added to the links output queue. Otherwise false if the send operation fails (e.g. if there is no link from this processor to one with the given processor ID).
     * @see #send(dans.algorithm.Message) 
     * @see #send(java.lang.String, java.lang.String) 
     * @see #sendLink(java.lang.String, java.lang.String) 
     * @see #sendLink(java.lang.String, java.lang.String, int) 
     */
    public final boolean send(String id, String msg, int port) {
        BackEndMessage bm = new BackEndMessage();
        bm.msg = msg;
        bm.fromID = getID();
        bm.toID = id;
        bm.port = getPort();
        bm.senderPort = getPort();

        if(port >= 0) {
           bm.port = port; 
        }
        
        return p.send(ns, bm, this);
    }
    
    
    /**
     * Sends the given message down the link with the given ID.
     * 
     * This method does NOT block or wait for the message to be delivered. Depending
     * on the failure settings delivery is not necessarily guaranteed. This method
     * returning true does not necessarily guarantee the message was delivered.
     * 
     * @param linkID the ID of the link to send the message down.
     * @param msg the message to send down the link.
     * @return true if the message was successfully added to the links output queue. Otherwise false if the send operation fails (e.g. if there is no link from this processor with the given link ID).
     * @see #send(dans.algorithm.Message) 
     * @see #send(java.lang.String, java.lang.String) 
     * @see #send(java.lang.String, java.lang.String, int) 
     * @see #sendLink(java.lang.String, java.lang.String, int) 
     */
    public final boolean sendLink(String linkID, String msg) {
        return sendLink(linkID, msg, -1);
    }
    
    /**
     * Sends the given message down the link with the given ID on the given port.
     * Specifying a port allows this algorithm to send messages to algorithms running 
     * on different ports on different processors.
     * 
     * This method does NOT block or wait for the message to be delivered. Depending
     * on the failure settings delivery is not necessarily guaranteed. This method
     * returning true does not necessarily guarantee the message was delivered.
     * 
     * @param linkID the ID of the link to send the message down.
     * @param msg the message to send down the link.
     * @param port the port to send the message to on the receiving processor.
     * @return true if the message was successfully added to the links output queue. Otherwise false if the send operation fails (e.g. if there is no link from this processor with the given link ID).
     * @see #send(dans.algorithm.Message) 
     * @see #send(java.lang.String, java.lang.String) 
     * @see #send(java.lang.String, java.lang.String, int) 
     * @see #sendLink(java.lang.String, java.lang.String) 
     */
    public final boolean sendLink(String linkID, String msg, int port) {
        BackEndMessage bm = new BackEndMessage();
        bm.msg = msg;
        bm.fromID = getID();
        bm.linkID = linkID;
        bm.port = getPort();
        bm.senderPort = getPort();
        
        if(port >= 0) {
           bm.port = port; 
        }
        
        return p.send(ns, bm, this);
    }
    
    /**
     * Returns the next message in the message buffer (the message input queue).
     * 
     * This method does not block or wait for messages and returns immediately.
     * At most one message will be returned with each call to receive.
     * 
     * @return a message sent to this processor/algorithm or null if the message buffer is empty (no more messages to get).
     * @see Message
     * @see #receive(boolean) 
     */
    public final Message receive() {
        return receive(true);
    }
    
    /**
     * Returns the next message in the message buffer (the message input queue).
     * 
     * Depending on the value given to this method it will either return immediately
     * or block/wait until a message is received. If the method waits, it will sleep
     * and disregard the tick system. There is no guarantee the algorithm will ever
     * wake before it is terminated (e.g. if no message is sent to the processor).
     * After the algorithm wakes and a message is returned it will resume running 
     * at a speed dictated by the tick system.
     * 
     * @param asynchronous if true the method will return immediately and function like {@link #receive()}. If false the method will block/wait until a message is received.
     * @return a message sent to this processor/algorithm or null if the message buffer is empty (no more messages to get).
     * @see Message
     * @see #receive()
     */
    public final Message receive(boolean asynchronous) {
        MessageInQueue in = p.getQueue(getPort());
        
        if(in != null) {
            BackEndMessage bm;
            
            if(asynchronous) {
                bm = in.receive();
            } else {
                synchronized(statusLock) {
                    status = NetworkObject.STATUS_WAITING;
                    p.algorithmUpdateStatus(this, status);
                }
                bm = in.receiveWait();
                synchronized(statusLock) {
                    status = NetworkObject.STATUS_RUNNING;
                    p.algorithmUpdateStatus(this, status);
                }
            }
            
            if(bm != null) {
                return bm.makeMessage();
            }
        }
        
        return null;
    }
    
    /**
     * This method should be checked once per cycle of the main loop in the {@link #algorithm()}
     * method. The main loop should be terminated if this function returns false.
     * 
     * This method both ensures that the algorithm is running at the speed dictated
     * by the tick system and updates dictated about the algorithm. It also allows
     * the simulation to be paused/restarted and handles failures. See {@link Algorithm}
     * for more details and an example of its use.
     * 
     * @return false if the algorithm's main loop should be terminated. Otherwise true.
     * @see Algorithm
     */
    public final boolean doMainLoop() {
        if(!firstloop && !terminated) {
            synchronized(statusLock) {
                if(status.equals(NetworkObject.STATUS_RUNNING)) {
                    status = NetworkObject.STATUS_WAITING;
                    p.algorithmUpdateStatus(this, status);
                }
            }
            
            int oldTick = ns.getTick();

            int delay = Utils.makeInt(p.getSetting("Delay"));

            synchronized(commonLock) {
                while((oldTick + delay > ns.getTick() || paused) && !terminated) {
                    try {
                        commonLock.wait(THREAD_RETRY_TIME);
                    } catch (InterruptedException ex) {}
                }
            }
            
            if(!terminated) {
                if(delay <= 0) {
                    try {
                        sleep(THREAD_MIN_WAIT_TIME);
                    } catch (InterruptedException ex) {}
                }

                ns.incrementNetworkStat("Cycles", this);
                p.incrementStat("Cycles", this);


                synchronized(statusLock) {
                    status = NetworkObject.STATUS_RUNNING;
                    p.algorithmUpdateStatus(this, status);
                }
            }
        } else {
            firstloop = false;
        }
        
        return !terminated && !checkCleanFails();
    }
    
    /**
     * Causes this algorithm to terminate cleanly and for {@link #doMainLoop()} to
     * return false the next time it is called. Commands after calling terminate
     * will still be executed.
     */
    public final void terminate() {
        terminated = true;
    }

    /**
     * This method causes the algorithm to be unpaused. This method should not be
     * called by a child of the Algorithm class.
     */
    public final void unpause() {
        paused = false;
        synchronized(statusLock) {
            status = NetworkObject.STATUS_WAITING;
            p.algorithmUpdateStatus(this, status);
        }
    }

    /**
     * This method causes the algorithm to be paused. This method should not be
     * called by a child of the Algorithm class.
     */
    public final void pause() {
        paused = true;
        synchronized(statusLock) {
            status = NetworkObject.STATUS_PAUSED;
            p.algorithmUpdateStatus(this, status);
        }
    }
    
    /**
     * Checks if this algorithm should fail based on the clean failure settings
     * for the processor it is running on. Each time this method is called there
     * is a chance the algorithm will fail based on the clean failure probability
     * set in the processor. This method should not be called from a child of the
     * Algorithm class.
     * 
     * @return true if the algorithm failed. Otherwise false.
     * @see #fail() 
     * @see #fail(boolean) 
     */
    private final boolean checkCleanFails() {
        boolean docheck = false;
        synchronized(statusLock) {
            if(status.equals(NetworkObject.STATUS_RUNNING) || status.equals(NetworkObject.STATUS_WAITING)) {
                docheck = true;
            }
        }
        
        Object cleanFailRate = p.getSetting("Clean Failure Rate");
        if(docheck && cleanFailRate != null) {
            double failRate = Utils.makeDouble(cleanFailRate);
            double roll = rand.nextDouble();
            
            if(failRate > 0 && roll < (failRate/100d)) {
                fail();
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Causes the algorithm to experience a clean failure. All messages will be
     * sent and then the algorithm will be terminated. This should not be called
     * to simply terminate the algorithm, the {@link #terminate()} method should
     * be used for that purpose. This method should only be used to simulate a
     * clean failure.
     * 
     * @see #terminate() 
     */
    public final void fail() {
        fail(true);
    }
    
    /**
     * Causes the algorithm to experience a failure. The type of failure depends
     * on the value passed to the method. If a clean failure is chosen all messages 
     * will be sent and then the algorithm will be terminated. If a unclean failure
     * is chosen no guarantee is made that this method will send all messages before
     * it terminates.
     * 
     * This method should not be called by a child of the Algorithm class.
     * 
     * @param clean true if it is a clean failure. Otherwise false.
     */
    public final void fail(boolean clean) {
        synchronized(failLock) {
            if(failType == FAIL_NONE) {

                if(clean) {
                    ns.incrementNetworkStat("Clean Fails", this);
                } else {
                    ns.incrementNetworkStat("Byzantine Fails", this);
                }

                synchronized(statusLock) {
                    status = NetworkObject.STATUS_FAILED;
                    p.algorithmUpdateStatus(this, status);
                }

                paused = false;
                terminate();

                if(clean) {
                    failType = FAIL_CLEAN;
                    Log.info("Algorithm encountered clean failure.", "P:" + p.getID() + ":" + getPort());
                } else {
                    failType = FAIL_BYZAN;
                    Log.info("Algorithm encountered byzantine failure.", "P:" + p.getID() + ":" + getPort());
                } 
            }
        }
    }
}
