package dans.algorithm;

import dans.util.Utils;

/**
 * The Message class warps a message sent between processors with meta data about
 * how it is or will be transmitted. To create a message to be sent, one of the
 * constructors should be used. Received Message objects can not be resent, instead
 * a new message object with the same message should be created. After a message
 * object is created, none of its values can be changed (this is intentional).
 * 
 * All messages are treated as strings, however, other values can be encoded in them.
 * If nonprintable characters are used or the message is long, it may not be displayed
 * correctly in the front-end GUI.
 * 
 * @author Daniel Servos
 * @version 0.1beta
 */
public final class Message {
    private String fromID = null;
    private String toID = null;
    private String linkID = null;
    private String msg = null;
    private int sentTick = -1;
    private int receivedTick = -1;
    private int port = -1;
    private int sport = -1;
    
    
    private final void setUp(String fromID, String toID, String linkID, String msg, Integer sentTick, Integer receivedTick, Integer port, Integer sport) {
        this.fromID = fromID;
        this.toID = toID;
        this.linkID = linkID;
        this.msg = msg;
        this.sentTick = sentTick;
        this.receivedTick = receivedTick;
        this.port = port;
        this.sport = sport;
    }
    
    /**
     * Creates a new Message object to be sent to a given processor.
     * 
     * Message objects can be sent using the {@link Algorithm#send(dans.algorithm.Message)}
     * command in the {@link Algorithm} class.
     * 
     * @param toID the processor's ID
     * @param message the message to be sent.
     * @see Algorithm#send(dans.algorithm.Message)
     * @see #Message(java.lang.String, java.lang.String, boolean) 
     * @see #Message(java.lang.String, java.lang.String, int) 
     * @see #Message(java.lang.String, java.lang.String, int, boolean) 
     */
    public Message(String toID, String message) {
        this(toID, message, false);
    }
    
    /**
     * Creates a new Message object to be set to a given processor on a given port.
     * 
     * Message objects can be sent using the {@link Algorithm#send(dans.algorithm.Message)}
     * command in the {@link Algorithm} class.
     * 
     * @param toID the processor's ID
     * @param message the message to be sent.
     * @param port the port to send the message to on the target processor.
     * 
     * @see Algorithm#send(dans.algorithm.Message)
     * @see #Message(java.lang.String, java.lang.String) 
     * @see #Message(java.lang.String, java.lang.String, boolean) 
     * @see #Message(java.lang.String, java.lang.String, int, boolean) 
     */
    public Message(String toID, String message, int port) {
        this(toID, message, port, false);
    }
    
    /**
     * Creates a new Message object to be sent to a given processor or link. If the
     * isLinkID parameter is set to true, toID is assumed to be a link ID and the
     * message will be sent down the given link when sent. If isLinkID is false,
     * toID is assumed to be a processor ID and the message will be sent to the
     * given processor when sent.
     * 
     * Message objects can be sent using the {@link Algorithm#send(dans.algorithm.Message)}
     * command in the {@link Algorithm} class.
     * 
     * @param toID the ID of the processor or link the message is for.
     * @param message the message to be sent.
     * @param isLinkID if toID is a link ID or a processor ID.
     * @see Algorithm#send(dans.algorithm.Message)
     * @see #Message(java.lang.String, java.lang.String) 
     * @see #Message(java.lang.String, java.lang.String, int) 
     * @see #Message(java.lang.String, java.lang.String, int, boolean) 
     */
    public Message(String toID, String message, boolean isLinkID) {
        msg = message;
        if(isLinkID) {
            this.linkID = toID;
        } else {
            this.toID = toID;
        }
    }
    
    /**
     * Creates a new Message object to be sent to a given processor or link on a
     * given port. If the isLinkID parameter is set to true, toID is assumed to be a link ID and the
     * message will be sent down the given link when sent. If isLinkID is false,
     * toID is assumed to be a processor ID and the message will be sent to the
     * given processor when sent.
     * 
     * Message objects can be sent using the {@link Algorithm#send(dans.algorithm.Message)}
     * command in the {@link Algorithm} class.
     * 
     * @param toID the ID of the processor or link the message is for.
     * @param message the message to be sent.
     * @param port the port to send the message to on the target processor.
     * @param isLinkID if toID is a link ID or a processor ID.
     */
    public Message(String toID, String message, int port, boolean isLinkID) {
        msg = message;
        this.port = port;
        if(isLinkID) {
            this.linkID = toID;
        } else {
            this.toID = toID;
        }
    }
    
    /**
     * Gets time (in ticks) that the message was sent at or -1 if the message has
     * not yet been sent.
     * 
     * @return the time he message was sent at or -1 if the message has not yet been sent.
     */
    public final int sent() {
        return sentTick;
    }
    
    /**
     * Gets time (in ticks) that the message was received at or -1 if the message has
     * not yet been received.
     * 
     * @return time that the message was received at or -1 if the message has not yet been received.
     */
    public final int received() {
        return receivedTick;
    }
    
    /**
     * Gets the processor ID of the processor that sent the message or null if the
     * message has not yet been sent.
     * 
     * @return a processor ID or null.
     */
    public final String from() {
        return fromID;
    }
    
    /**
     * Gets the processor ID of the processor that message was or will be sent to
     * or null if no processor ID is set (e.g. if the message will be sent by using
     * a link ID).
     * 
     * @return a processor ID or null.
     */
    public final String to() {
        return toID;
    }
    
    /**
     * Gets the link ID of the link that transmitted or will transmit the message
     * or null if no link ID is set (e.g. if the message has not yet been sent
     * and will be sent by using a processor ID).
     * 
     * @return a link ID or null.
     */
    public final String link() {
        return linkID;
    }
    
    /**
     * Gets the message as a string.
     * 
     * @return the message.
     */
    public final String message() {
        return msg;
    }
    
    /**
     * Gets the size of the message in bytes. This is equivalent to the length
     * of the string. The '\0' character is not counted in this size.
     * 
     * @return the size of the message in bytes. 
     */
    public final int size() {
        return msg.length();
    }
    
    /**
     * The port the algorithm was listening on that sent this message or -1 if the
     * message has not yet been sent. This allows an algorithm to respond to messages
     * on the correct port.
     * 
     * @return a port number or -1.
     * @see #port()
     */
    public final int senderPort() {
        return sport;
    }
    
    /**
     * Equivalent to {@link #message()}.
     * 
     * @return the message as a string.
     * @see #message()
     * @see #toInt() 
     * @see #toDouble() 
     * @see #toLong()
     */
    @Override
    public final String toString() {
        return msg;
    }
    
    /**
     * Attempts to convert the message string into an integer.
     * 
     * @return the message represented as an integer
     * @throws NumberFormatException if the string does not contain a parsable integer.
     * @see #toString() 
     * @see #toDouble() 
     * @see #toLong() 
     */
    public final int toInt() {
        return Utils.makeInt(msg);
    }
    
    /**
     * Attempts to convert the message string into a double.
     * 
     * @return the message represented as a double
     * @throws NumberFormatException if the string does not contain a parsable double.
     * @see #toString() 
     * @see #toInt() 
     * @see #toLong() 
     */
    public final double toDouble() {
        return Utils.makeDouble(msg);
    }

    /**
     * Attempts to convert the message string into a long.
     * 
     * @return the message represented as a long
     * @throws NumberFormatException if the string does not contain a parsable long.
     * @see #toString() 
     * @see #toInt() 
     * @see #toDouble() 
     */
    public final long toLong() {
        return Utils.makeLong(msg);
    }
    
    /**
     * Gets the port this message was sent to or will be sent to. -1 is returned
     * if this message has not yet been sent and the port is not set in the constructor.
     * 
     * @return the port number or -1.
     * @see #senderPort() 
     */
    public final int port() {
        return port;
    }
    
    /**
     * Duplicate of the {@link #senderPort()} method. To be removed.
     * 
     * @return the port number or -1.
     * @deprecated duplicate of {@link #senderPort()}.
     * @see #senderPort()
     */
    public final int sourcePort() {
        return sport;
    }
}
