package dans.network;

import dans.util.Log;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.concurrent.ConcurrentHashMap;

/**
 *
 * @author Dan
 */
public class Link extends NetworkObject {
    public static final int EVENT_TARGET_UPDATE = 100;
    public static final int EVENT_SOURCE_UPDATE = 101;
    
    public static final int EVENT_MESSAGE_SENT = 110;
    public static final int EVENT_MESSAGE_TRANSFERRED = 111;
    
    private Processor source;
    transient private Object sourceLock = new Object();
    private Processor target;
    transient private Object targetLock = new Object();

    transient private MessageOutQueue queue = null;
    
    public Link(String id) {
        this(id, null, null);
    }
    
    public Link(String id, Processor source, Processor target) {
        super(id);
        this.source = source;
        this.target = target;
    }
    
    public Link(String id, Link l) {
        super(id);
        copyFrom(l);
    }
    
    public Processor getSource() {
        return source;
    }
    
    public Processor getTarget() {
        return target;
    }
    
    public boolean isComplete(){
        return target != null && source != null;
    }
    
    public void setSource(Processor p) {
        setSource(p, null);
    }
    
    public void setTarget(Processor p) {
        setTarget(p, null);
    }
    
    public void setSource(Processor p, Object caller) {
        Processor old = null;
        boolean change = false;
        
        synchronized(sourceLock) {
            if(p != source) {
                old = source;
                source = p;
                change = true;
            }
        }

        if(change) {
            notifyListeners(old, EVENT_SOURCE_UPDATE, caller);
        }
    }
    
    public void setTarget(Processor p, Object caller) {
        Processor old = null;
        boolean change = false;
        
        synchronized(targetLock) {
            if(p != target) {
                old = target;
                target = p;
                change = true;
            }
        }

        if(change) {
            notifyListeners(old, EVENT_TARGET_UPDATE, caller);
        }
    }

    @Override
    public void setDefaultSettings(NetworkState ns) {
        synchronized(settingsLock) {
            settings.put("Delay", ns.getNetworkSetting("defaultLinkSpeed"));
            settings.put("Speed Change Method", ns.getNetworkSetting("defaultLinkSpeedChangeMethod"));
            settings.put("Change Min", ns.getNetworkSetting("defaultLinkSpeedChangeMin"));
            settings.put("Change Max", ns.getNetworkSetting("defaultLinkSpeedChangeMax"));
            settings.put("Normal Change Mean", ns.getNetworkSetting("defaultLinkSpeedChangeNormMean"));
            settings.put("Normal Change STDV", ns.getNetworkSetting("defaultLinkSpeedChangeNormSTDV"));
            settings.put("Uniform Change Min", ns.getNetworkSetting("defaultLinkSpeedChangeUniMin"));
            settings.put("Uniform Change Max", ns.getNetworkSetting("defaultLinkSpeedChangeUniMax"));
            settings.put("Constant Change Amount", ns.getNetworkSetting("defaultLinkSpeedChangeConstant"));
            settings.put("Failure Rate", ns.getNetworkSetting("defaultLinkFailRate"));
            settings.put("Packet Loss Rate", ns.getNetworkSetting("defaultLinkLossRate"));
            settings.put("Byte Error Rate", ns.getNetworkSetting("defaultLinkErrorRate"));
            settings.put("Bandwidth", ns.getNetworkSetting("defaultLinkBandwidth"));
            settings.put("Change Seed", ns.getNetworkSetting("defaultSpeedChangeSeed"));
        }
    }

    @Override
    public void setDefaultStats(NetworkState ns) {
        synchronized(statsLock) {
            stats.put("Messages", 0);
            stats.put("Messages in Bytes", 0);
            stats.put("Messages per Tick", 0);
            stats.put("Bytes per Tick", 0);
            stats.put("Packets Lost", 0);
            stats.put("Packet Errors", 0);
        }
    }

    @Override
    public String getType() {
        return "Link";
    }
    
    public void setQueue(MessageOutQueue queue) {
        this.queue = queue;
    }
    
    public boolean send(BackEndMessage bm) {
        String status = getStatus();
        if(queue != null && !status.equals(NetworkObject.STATUS_FAILED) && !status.equals(NetworkObject.STATUS_ERROR) && !status.equals(NetworkObject.STATUS_TERMINATED)) {
            BackEndMessage clone = new BackEndMessage(bm);
            queue.send(bm);
            notifyListeners(clone, EVENT_MESSAGE_SENT, this);
            return true;
        }

        return false;
    }
    
    /*@Override
    protected void copyFrom(NetworkObject o) {
        super.copyFrom(o);
        Link l = (Link)o;
        synchronized(sourceLock) {
            synchronized(l.sourceLock) {
                source = l.source;
            }
        }
        synchronized(targetLock) {
            synchronized(l.targetLock) {
                target = l.target;
            }
        }
    }*/

    public void transfer() {
        String status = getStatus();
        if(queue != null && (status.equals(NetworkObject.STATUS_RUNNING) || status.equals(NetworkObject.STATUS_WAITING))) {
            boolean sent = queue.transfer();
            if(sent) {
                notifyListeners(queue.getLastOut(), EVENT_MESSAGE_TRANSFERRED, this);
            }
        }
    }
    
    public MessageOutQueue getQueue() {
        return queue;
    }
    
    public char getShortType() {
        return 'L';
    }
    
    @Override
    public void destory() {
        synchronized(sourceLock) {
            source = null;
        }
        
        synchronized(targetLock) {
            target = null;
        }

        queue = null;

        super.destory();
    }
    
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        resetForLoad();
        sourceLock = new Object();
        targetLock = new Object();
        queue = null;
    }
}
