package dans.network;

import dans.algorithm.Algorithm;
import dans.algorithm.Message;
import dans.util.Log;
import dans.util.Utils;
import java.util.Random;
import java.util.concurrent.ConcurrentLinkedQueue;

/**
 *
 * @author Dan
 */
public class MessageOutQueue {
    private Link link;
    private NetworkState state;
    private final Object lock = new Object();
    private final ConcurrentLinkedQueue<BackEndMessage> out = new ConcurrentLinkedQueue<>();
    
    private int messagesSent = 0;
    private int bytesSent = 0;
    
    private BackEndMessage lastOut = null;
    private int lastTransfer = -1;
    
    private boolean failed = false;
    
    private Random rand;
    
    
    public MessageOutQueue(NetworkState state, Link l) {
        this.state = state;
        link = l;
        rand = new Random();
    }
    
    private int delay() {
        return Utils.makeInt(link.getSetting("Delay"));
    }
    
    private double bandwith() {
        return Utils.makeDouble(link.getSetting("Bandwidth"));
    }
    
    public void send(BackEndMessage m) {
        if(checkSourceFail(m.senderPort)) {
            synchronized(lock) {
                m.sentTick = state.getTick();
                out.add(m);
            }

            Log.info("Sending message \"" + m + "\" from " + m.fromID + " to " + m.toID, "P:" + m.fromID + "#" + m.port);

            if(delay() <= 0) {
                link.transfer();
            }
        }
    }
    
    private boolean checkSourceFail(int port) {
        Processor s = link.getSource();
        Object failRate = s.getSetting("Byzantine Failure Rate");
        int failType = s.getFailType(port);
        if(failRate != null && failType == Algorithm.FAIL_NONE) {
            double rate = Utils.makeDouble(failRate);
            double roll = rand.nextDouble();
            
            if(rate > 0 && roll < (rate/100d)) {
                s.failAlgorithm(port, false);
                return true;
            }
        }
        
        return failType == Algorithm.FAIL_NONE || failType == Algorithm.FAIL_CLEAN;
    }
    
    private String messageErrors(String message) {
        Object errorRate = link.getSetting("Byte Error Rate");
        StringBuilder sb = new StringBuilder();
        
        if(message != null && errorRate != null) {
            double rate = Utils.makeDouble(errorRate);
            for(char c : message.toCharArray()) {
                double roll = rand.nextDouble();
                if(rate > 0 && roll < (rate/100d)) {
                    state.incrementNetworkStat("Link Errors", link);
                    link.incrementStat("Packet Errors", link);
                    sb.append(Character.toChars(rand.nextInt(95) + 32));
                    Log.debug("Byte error.", "L:" + link.getID());
                } else {
                    sb.append(c);
                }
            }
        } else {
            return message;
        }
        
        return sb.toString();
    }
    
    private boolean checkPacketLoss() {
        Object lossRate = link.getSetting("Packet Loss Rate");
        if(lossRate != null) {
            double rate = Utils.makeDouble(lossRate);
            double roll = rand.nextDouble();
            
            if(rate > 0 && roll < (rate/100d)) {
                state.incrementNetworkStat("Packets Lost", link);
                link.incrementStat("Packets Lost", link);
                Log.debug("Packet lost.", "L:" + link.getID());
                return false;
            }
        }
        
        return true;
    }
    
    private boolean checkLinkFail() {
        Object failRate = link.getSetting("Failure Rate");
        
        if(failRate != null && !failed) {
            double rate = Utils.makeDouble(failRate);
            double roll = rand.nextDouble();
            
            if(rate > 0 && roll < (rate/100d)) {
                failed = true;
                state.incrementNetworkStat("Link Fails", link);
                link.setStatus(NetworkObject.STATUS_FAILED, link);
                Log.debug("Link failed", "L:" + link.getID());
            }
        }
        
        return !failed;
    }
    
    public boolean transfer() {
        boolean keepgoing = false;
        
        synchronized(lock) {
            if(checkLinkFail()) {
                BackEndMessage head = out.peek();

                int tick = state.getTick();
                double bwidth = bandwith();
                double current = (double)(messagesSent+1)/(double)tick;
                
                if(bwidth <= 0 || current <= bwidth) {
                    if(head != null && head.sentTick + delay() <= tick) {
                        lastOut = out.poll();
                        if(checkPacketLoss()) {
                            head.msg = messageErrors(head.msg);
                            link.getTarget().receive(link, head);
                            link.incrementStat("Messages", link);
                            link.incrementStat("Messages in Bytes", head.size(), link);
                            messagesSent++;
                            bytesSent += head.size();
                            link.setStat("Messages per Tick", Utils.round((double)messagesSent/(double)tick, 2));
                            link.setStat("Bytes per Tick", Utils.round((double)bytesSent/(double)tick, 2));
                        }
                        lastTransfer = tick;
                        keepgoing = true;
                    }
                }
            }

            if(keepgoing) {
                transfer();
            }
        }
        
        return keepgoing;
    }

    public int getSize() {
        synchronized(lock) {
            return out.size();
        }
    }
    
    public BackEndMessage peek() {
        synchronized(lock) {
            return out.peek();
        }
    }
    
    public BackEndMessage getLastOut() {
        synchronized(lock) {
            return lastOut;
        }
    }
    
    public int getLastTransfer() {
        synchronized(lock) {
            return lastTransfer;
        }
    }
}
