package dans.network;

import dans.algorithm.Algorithm;
import dans.util.Log;
import dans.util.Utils;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author Dan
 */
public class NetworkManager implements NetworkObjectListener, NetworkStateListener {
    private NetworkState state;
    private int lastPID = 0;
    private int lastLID = 0;
    private final Object lastPIDLock = new Object();
    private final Object lastLIDLock = new Object();
    
    private boolean started = false;
    private final Object startedLock = new Object();
    
    
    private List<Algorithm> algorithms = Collections.synchronizedList(new ArrayList<>());
    
    private Object commonLock;
    
    
    private final int THREAD_WAIT_TIME = 25;
    private final int THREAD_PAUSE_WAIT_TIME = 250;
    
    private boolean paused = false;
    
    public NetworkManager(NetworkState state) {
        this.state = state;
        this.commonLock = state.getCommonLock();
        state.addListener(this, false);
        state.addListenerToAllProcessors(this, false);
        state.addListenerToAllLinks(this, false);
    }

    public void startUp() {
        Log.debug("Main thread waiting for network simulation start.");
        synchronized(startedLock) {
            while(!started) {
                try {
                    startedLock.wait();
                } catch (InterruptedException ex) {}
            }
        }
        
        Log.debug("Initializing algorithms.");
        
        Processor procs[] = state.getProcessors();
        for(Processor p : procs) {
            p.initializeAlgorithms(state);
            algorithms.addAll(p.getAlgorithmInstance());
        }
        
        Log.debug("Initializing message queues.");
        
        Link links[] = state.getLinks();
        for(Link l : links) {
            l.setQueue(new MessageOutQueue(state, l));
            l.setStatus(NetworkObject.STATUS_RUNNING, this);
        }
        
        
        Log.debug("Starting network simulation.");
        
        for(Processor p : procs) {
            if(p.getStatus().equals(NetworkObject.STATUS_NOTSTARTED)) {
                p.startAlgorithms();
            }
        }
    }
    
    private void transferMessages() {
        Link links[] = state.getLinks();
        for(Link l : links) {
            l.transfer();
        }
    }
    
    private void doSpeedChanges() {
        Link links[] = state.getLinks();
        for(Link l : links) {
            if(l.getStatus().equals(NetworkObject.STATUS_RUNNING) || l.getStatus().equals(NetworkObject.STATUS_WAITING) || l.getStatus().equals(NetworkObject.STATUS_PAUSED)) {
                l.doSpeedChange();
            }
        }
        
        Processor procs[] = state.getProcessors();
        for(Processor p : procs) {
            if(p.getStatus().equals(NetworkObject.STATUS_RUNNING) || p.getStatus().equals(NetworkObject.STATUS_WAITING) || p.getStatus().equals(NetworkObject.STATUS_PAUSED)) {
                p.doSpeedChange();
            }
        }
    }
    
    private boolean checkRunning() {
        int termed = 0;
        int running = 0;
        
        Processor procs[] = state.getProcessors();
        for(Processor p : procs) {
            if(p.getStatus().equals(NetworkObject.STATUS_TERMINATED) || p.getStatus().equals(NetworkObject.STATUS_ERROR) || p.getStatus().equals(NetworkObject.STATUS_FAILED)) {
                termed++;
            } else {
                running++;
            }
        }
        
        return running > 0 && started;
    }
    
    private void finalStatsUpdate() {
        Processor procs[] = state.getProcessors();
        for(Processor p : procs) {
            String status = p.checkTermStatus();
            p.setStatus(status);
        }
        
        Link links[] = state.getLinks();
        for(Link l : links) {
            if(l.getStatus().equals(NetworkObject.STATUS_RUNNING) || l.getStatus().equals(NetworkObject.STATUS_WAITING) || l.getStatus().equals(NetworkObject.STATUS_PAUSED)) {
                l.setStatus(NetworkObject.STATUS_TERMINATED);
            }
        }
    }
    
    private void resetSettings() {
        Processor procs[] = state.getProcessors();
        for(Processor p : procs) {
            p.setDefaultSettings(state);
        }
        
        Link links[] = state.getLinks();
        for(Link l : links) {
            l.setDefaultSettings(state);
        }
    }
    
    private void resetStats() {
        Processor procs[] = state.getProcessors();
        for(Processor p : procs) {
            p.setDefaultStats(state);
        }
        
        Link links[] = state.getLinks();
        for(Link l : links) {
            l.setDefaultStats(state);
        }
        
        state.resetStats();
    }
    
    private void setAllStopped() {
        Processor procs[] = state.getProcessors();
        for(Processor p : procs) {
            p.setStatus(NetworkObject.STATUS_NOTSTARTED, this);
        }
        
        Link links[] = state.getLinks();
        for(Link l : links) {
            l.setStatus(NetworkObject.STATUS_NOTSTARTED, this);
        }
    }
    
    private void clear(Object caller) {
        resetStats();
        resetSettings();
        state.resetStats();
        state.resetSettings();
        state.clearProcessors();
        state.clearLinks();
        state.clearAlgorithms();
        
        synchronized(lastPIDLock) {
            lastPID = 0;
        }
        synchronized(lastLIDLock) {
            lastLID = 0;
        }
        synchronized(startedLock) {
            started = false;
            paused = false;
        }
        
        state.updateNetworkStat("Network State", SettingsConstants.STATE_STOPPED, caller);
    }
    
    private void load(NetworkState newstate) {
        newstate.addListener(this, false);
        newstate.addListenerToAllProcessors(this, false);
        newstate.addListenerToAllLinks(this, false);
        commonLock = newstate.getCommonLock();
        state = newstate;
    }
    
    private void reset() {
        Log.info("Resetting simulation.");
        synchronized(startedLock) {
            if(started) {
                started = false;
            }
        }
        
        Log.debug("Interrupting algorithm threads.");
        synchronized(commonLock) {
            synchronized(algorithms){
                for(Algorithm a : algorithms) {
                    a.unpause();
                    a.terminate();
                    a.interrupt();
                }
            }
            
            paused = false;
            commonLock.notifyAll();
        }
        
        Log.debug("Forcing algorithm threads to end.");
        synchronized(algorithms){
            for(Algorithm a : algorithms) {
                try {
                    a.join(250);                    
                } catch (InterruptedException ex) {}
                
                                    
                if(a.getState() != Thread.State.TERMINATED){
                    Log.debug("Had to kill " + a.getID() + "#" + a.getPort());
                    a.stop();
                }
            }
            
            algorithms.clear();
        }
        
        
        Log.debug("Resetting settings and stats.");
        resetStats();
        //resetSettings();
        setAllStopped();
        Log.debug("Simulation restarted.");
    }
    
    public void runManager() {
        Log.debug("Starting network manager main loop.");
                
        while(true){
            if(!state.getNetworkStat("Network State").equals(SettingsConstants.STATE_RUNNING)) {
                state.updateNetworkStat("Network State", SettingsConstants.STATE_RUNNING, this);
            }
            
            int algoReady = 0;
            boolean waiting;

            while(checkRunning()) {
                try {
                    int delay = Utils.makeInt(state.getNetworkSetting("tickSpeed"));
                    Thread.sleep(delay);
                } catch (InterruptedException ex) {
                   Log.debug("Manager main thread sleep interrupted.");
                }

                waiting = true;
                synchronized(commonLock) {
                    while(waiting && started) {
                        if(!paused && started) {
                            transferMessages();

                            algoReady = 0;
                            synchronized(algorithms){
                                for(Algorithm a : algorithms) {
                                    if(!a.getStatus().equals(NetworkObject.STATUS_RUNNING)) {
                                        algoReady++;
                                    }
                                }
                            }
                        }
                        
                        if(algoReady < algorithms.size() || paused) {
                            try {
                                commonLock.wait(THREAD_WAIT_TIME);
                            } catch (InterruptedException ex) {}
                        } else if(started) {
                            state.incrementTick(this);
                            transferMessages();
                            doSpeedChanges();
                            commonLock.notifyAll();
                            waiting = false;
                        }
                    }
                }
            }

            finalStatsUpdate();
            if(started) {
                state.updateNetworkStat("Network State", SettingsConstants.STATE_STOPPED, this);
            }
            
            synchronized(startedLock) {
                started = false;
            }
            startUp();
        }
    }
    
    public void pause() {
        synchronized(algorithms){
            for(Algorithm a : algorithms) {
                if(a.getStatus().equals(NetworkObject.STATUS_RUNNING) || a.getStatus().equals(NetworkObject.STATUS_WAITING)) {
                    a.pause();
                }
            }
        }
        
        Link links[] = state.getLinks();
        for(Link l : links) {
            if(l.getStatus().equals(NetworkObject.STATUS_RUNNING) ||  l.getStatus().equals(NetworkObject.STATUS_WAITING)) {
                l.setStatus(NetworkObject.STATUS_PAUSED);
            }
        }
        
        Processor procs[] = state.getProcessors();
        for(Processor p : procs) {
            if(p.getStatus().equals(NetworkObject.STATUS_RUNNING) ||  p.getStatus().equals(NetworkObject.STATUS_WAITING)) {
                p.setStatus(NetworkObject.STATUS_PAUSED);
            }
        }
        
        synchronized(commonLock) {
            paused = true;
        }
    }
    
    public void unpause() {
        Link links[] = state.getLinks();
        for(Link l : links) {
            if(l.getStatus().equals(NetworkObject.STATUS_PAUSED)) {
                l.setStatus(NetworkObject.STATUS_RUNNING);
            }
        }
        
        Processor procs[] = state.getProcessors();
        for(Processor p : procs) {
            if(p.getStatus().equals(NetworkObject.STATUS_PAUSED)) {
                p.setStatus(NetworkObject.STATUS_RUNNING);
            }
        }
        
        synchronized(algorithms){
            for(Algorithm a : algorithms) {
                if(a.getStatus().equals(NetworkObject.STATUS_PAUSED) || a.getStatus().equals(NetworkObject.STATUS_RUNNING) || a.getStatus().equals(NetworkObject.STATUS_WAITING)) {
                    a.unpause();
                }
            }
        }
        
        synchronized(commonLock) {
            paused = false;
            commonLock.notifyAll();
        }
    }
    
    @Override
    public void networkObjectChange(NetworkObject object, Object oldvalue, int eventID, Object caller) {
        if(object instanceof Processor) {
            Processor p = (Processor)object;
            
            if(eventID == Processor.EVENT_ID_UPDATE) {
                if(!state.updateProcessorID(p, oldvalue.toString(), this)) {
                    int id;
                    synchronized(lastPIDLock) {
                        id = lastPID++;
                    }
                    p.setID("ID IN USE [" + id + "]", this);
                    state.updateProcessorID(p, oldvalue.toString());
                }
            }
        } else if(object instanceof Link) {
            Link l = (Link)object;

            if(eventID == Link.EVENT_ID_UPDATE) {
                if(!state.updateLinkID(l, oldvalue.toString(), this)) {
                    int id;
                    synchronized(lastLIDLock) {
                        id = lastLID++;
                    }
                    l.setID("ID IN USE [" + id + "]", this);
                    state.updateLinkID(l, oldvalue.toString());
                }
            } else if (eventID == Link.EVENT_SOURCE_UPDATE) {
                
                if(oldvalue != null) {
                    ((Processor)oldvalue).removeNeighbour(l, this);
                }
                
                if(((Link)object).getSource() != null) {
                    if(((Link)object).getTarget() != null) {
                        ((Link)object).getSource().addNeighbour(l, ((Link)object).getTarget(), this);
                    }
                    Log.debug("Link " + object.getID() + " updated its source to " + ((Link)object).getSource().getID(), "L:" + object.getID());
                } else {
                    Log.debug("Link " + object.getID() + " removed its source.", "L:" + object.getID());
                }
            } else if (eventID == Link.EVENT_TARGET_UPDATE) {
                if(((Link)object).getSource() != null) {
                    ((Link)object).getSource().removeNeighbour(l, this);
                }
                
                if(((Link)object).getTarget() != null) {
                    if(((Link)object).getSource() != null) {
                        ((Link)object).getSource().addNeighbour(l, ((Link)object).getTarget(), this);
                    }
                    
                    Log.debug("Link " + object.getID() + " updated its target to " + ((Link)object).getTarget().getID(), "L:" + object.getID());
                } else {
                    Log.debug("Link " + object.getID() + " removed its target.", "L:" + object.getID());
                }
            }
        }
    }

    @Override
    public void processorChange(Processor p, int eventID, Object caller) {
        if(eventID == NetworkState.EVENT_PROC_ADD) {
            p.addListener(this, false);
        } else if(eventID == NetworkState.EVENT_PROC_REMOVE) {
            p.removeListener(this);
        }
    }

    @Override
    public void linkChange(Link l, int eventID, Object caller) {
        if(eventID == NetworkState.EVENT_LINK_ADD) {
            l.addListener(this, false);
            
            if(l.isComplete()) {
                l.getSource().addNeighbour(l, l.getTarget(), this);
                l.getTarget().addSource(l, l.getSource(), this);
            }
        } else if(eventID == NetworkState.EVENT_LINK_REMOVE) {
            l.removeListener(this);
            if(l.isComplete()) {
                l.getSource().removeNeighbour(l, this);
                l.getTarget().removeSource(l,this);
            }
        }
    }

    @Override
    public void settingsChange(String name, Object value, int eventID, Object caller) {
    }

    @Override
    public void statsChange(String name, Object value, int eventID, Object caller) {
        if(eventID == NetworkState.EVENT_STAT_UPDATE && name.equals("Network State")) {
            if(value.equals(SettingsConstants.STATE_RUNNING)) {
                synchronized(startedLock) {
                    if(!started) {
                        started = true;
                        startedLock.notifyAll();
                    } else if(paused){
                        unpause();
                    }
                }
            } else if(value.equals(SettingsConstants.STATE_PAUSED) && !paused && started) {
                pause();
            } else if(value.equals(SettingsConstants.STATE_RESET)) {
                reset();
            } else if(value.equals(SettingsConstants.STATE_CLEAR)) {
                clear(caller);
            } else if(value.equals(SettingsConstants.STATE_LOAD)) {
                load((NetworkState)caller);
            }
        }      
    }

    @Override
    public void tickChange(int tickValue, int eventID, Object caller) {
        
    }
}
