package dans.network;

import dans.util.Log;
import dans.util.Utils;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Collections;
import java.util.IdentityHashMap;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ThreadLocalRandom;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.SwingUtilities;

/**
 *
 * @author Dan
 */
public abstract class NetworkObject implements Cloneable, Serializable {
    public static final int EVENT_ID_UPDATE = 1;
    public static final int EVENT_STATUS_UPDATE = 2;
    public static final int EVENT_SETTINGS_UPDATE = 3;
    public static final int EVENT_STATS_UPDATE = 4;
    
    public static final String STATUS_NOTSTARTED = "NOT STARTED";
    public static final String STATUS_WAITING = "WAITING";
    public static final String STATUS_FAILED = "FAILED";
    public static final String STATUS_TERMINATED = "TERMINATED";
    public static final String STATUS_PAUSED = "PAUSED";
    public static final String STATUS_RUNNING = "RUNNING";
    public static final String STATUS_ERROR = "ERROR";
    
    
    protected ConcurrentHashMap<String, Object> settings = new ConcurrentHashMap<>();
    protected ConcurrentHashMap<String, Object> stats = new ConcurrentHashMap<>();
    
    transient protected Object statsLock = new Object();
    transient protected Object settingsLock = new Object();
    
    transient protected String status = STATUS_NOTSTARTED;
    transient protected Object statusLock = new Object();
    
    private String id;
    transient private Object idLock = new Object();
    transient private Map<NetworkObjectListener, Boolean> listeners = Collections.synchronizedMap(new IdentityHashMap<>());
    
    transient private Random rand = null;
    
    public NetworkObject(String id) {
        this.id = id;
    }
    
    public String getStatus() {
        synchronized(statusLock) {
            return status;
        }
    }
    
    public void setStatus(String status) {
        setStatus(status, null);
    }
    
    public void setStatus(String status, Object caller) {
        boolean changed = false;
        String oldStatus;
        
        synchronized(statusLock) {
            oldStatus = this.status;
            if(!oldStatus.equals(status)) {
                changed = true;
                this.status = status;
            }
        }
        
        if(changed) {
            notifyListeners(oldStatus, NetworkObject.EVENT_STATUS_UPDATE, caller);
        }
    }
    
    public void incrementSetting(String name, Object caller) {
        incrementSetting(name, 1, caller);
    }
    
    public void incrementSetting(String name) {
        incrementSetting(name, 1, null);
    }
    
    public void incrementSetting(String name, int by) {
        incrementSetting(name, by, null);
    }
    
    public void incrementSetting(String name, int by, Object caller) {
        boolean changed = false;
        int value = -1;
        Object oldValue[] = new Object[2];
        
        synchronized(settingsLock) {
            boolean ok = false;
            oldValue[0] = name;
            Object setting = settings.get(name);
            oldValue[1] = setting;
           
            if(setting != null) {
                try {
                    if(setting instanceof String) {
                        value = Integer.parseInt((String)setting);
                    } else {
                        value = (int)setting;
                    }
                    ok = true;
                } catch(Exception e) {}
                
                if(ok) {
                    settings.put(name, (value + by));
                    changed = true;
                }
            }
        }
        
        if(changed) {
            notifyListeners(oldValue, NetworkObject.EVENT_SETTINGS_UPDATE, caller);
        }
    }
    
    public void setSetting(String name, Object value) {
        setSetting(name, value, null);
    }
    
    public void setSetting(String name, Object value, Object caller) {
        boolean changed = false;
        Object oldValue[] = new Object[2];
        
        synchronized(settingsLock) {
            oldValue[0] = name;
            oldValue[1] = settings.get(name);

            //TODO: In some cases this is still trigered if the value is the same.
            if(!value.equals(oldValue[1])) {
                settings.put(name, value);
                changed = true;
            }
        }
        
        if(changed) {
            Log.debug(this.toString() + " updated " + name + " to " + value.toString());
            notifyListeners(oldValue, NetworkObject.EVENT_SETTINGS_UPDATE, caller);
        }
    }
    
    public Object getSetting(String name) {
        synchronized(settingsLock) {
            return settings.get(name);
        }
    }
    
    public void setStat(String name, Object value) {
        setStat(name, value, null);
    }
    
    public void setStat(String name, Object value, Object caller) {
        boolean changed = false;
        Object oldValue = null;
        Object newValue[] = {name, value};
        
        synchronized(statsLock) {
            oldValue = stats.get(name);
            
            if(value != oldValue) {
                stats.put(name, value);
                changed = true;
            }
        }
        
        if(changed) {
            notifyListeners(newValue, NetworkObject.EVENT_STATS_UPDATE, caller);
        }
    }
    
    public void incrementStat(String name, Object caller) {
        incrementStat(name, 1, caller);
    }
    
    public void incrementStat(String name) {
        incrementStat(name, 1, null);
    }
    
    public void incrementStat(String name, int by) {
        incrementStat(name, by, null);
    }
    
    public void incrementStat(String name, int by, Object caller) {
        boolean changed = false;
        int value = -1;
        Object newValue[] = new Object[2];
        
        synchronized(statsLock) {
            boolean ok = false;
            newValue[0] = name;
            Object stat = stats.get(name);
           
            if(stat != null) {
                try {
                    if(stat instanceof String) {
                        value = Integer.parseInt((String)stat);
                    } else {
                        value = (int)stat;
                    }
                    ok = true;
                } catch(Exception e) {}
                
                if(ok) {
                    stats.put(name, (value + by));
                    newValue[1] = (value + by);
                    changed = true;
                }
            }
        }
        
        if(changed) {
            notifyListeners(newValue, NetworkObject.EVENT_STATS_UPDATE, caller);
        }
    }
    
    public Object getStat(String name) {
        synchronized(statsLock) {
            return stats.get(name);
        }
    }
    
    public abstract void setDefaultSettings(NetworkState ns);
    public abstract void setDefaultStats(NetworkState ns);
    
    public Object[][] dumpSettings(String keyOrder[]) {
        Object data[][];
        
        synchronized(settingsLock) {
            data = new Object[keyOrder.length][2];
            int i = 0;
            
            for(String name: keyOrder) {
                data[i][0] = name;
            
                if(name.equals("ID")) {
                    data[i][1] = getID();
                } else {
                    data[i][1] = settings.get(name);
                }
                    
                i++;
            }
        }
        
        return data;
    }
    
    public Object[][] dumpStats(String keyOrder[]) {
        Object data[][];
        
        synchronized(statsLock) {
            data = new Object[keyOrder.length][2];
            int i = 0;
            
            for(String name: keyOrder) {
                data[i][0] = name;
                
                if(name.equals("Status")) {
                    data[i][1] = getStatus();
                } else {
                    data[i][1] = stats.get(name);
                }
                
                i++;
            }
        }
        
        return data;
    }
    
    public String getID() {
        synchronized(idLock) {
            return id;
        }
    }
    
    public void setID(String id) {
        setID(id, null);
    }
    
    public void setID(String newid, Object caller) {
        boolean changed = false;
        String oldid = this.id;
        
        synchronized(idLock) {
            if(!newid.equals(this.id)) {
                this.id = newid;
                changed = true;
            }
        }
        
        if(changed) {
            notifyListeners(oldid, NetworkObject.EVENT_ID_UPDATE, caller);
        }
    }
    
    public void addListener(NetworkObjectListener listener) {
        addListener(listener, false);
    }

    public void addListener(NetworkObjectListener listener, boolean useInvokeLater) {
        synchronized(listeners) {
            listeners.put(listener, useInvokeLater);
        }
    }
    
    public void removeListener(NetworkObjectListener listener) {
        synchronized(listeners) {
            listeners.remove(listener);
        }
    }
    
    protected void notifyListeners(Object oldvalue, int eventID, Object caller) {
        synchronized(listeners) {   
            listeners.keySet().stream().forEach((listener) -> {
                if(listener != caller) {
                    if(listeners.get(listener)) {
                        NetworkObject tmp = this;
                        SwingUtilities.invokeLater(() -> {listener.networkObjectChange(tmp, oldvalue, eventID, caller);});
                        //SwingUtilities.invokeLater(() -> listener.networkObjectChange(this, oldvalue, eventID, caller));
                    } else {
                        listener.networkObjectChange(this, oldvalue, eventID, caller);
                    }
                }
            });
        }
    }
    
    public abstract String getType();
    
    @Override
    public String toString() {
        synchronized(idLock) {
            return getType() + " " + this.id;
        }
    }
    
    public abstract char getShortType();
    
    public String toLogID() {
        return getShortType() + ":" + getID();
    }
    
    protected void copyFrom(NetworkObject o) {
        synchronized(settingsLock) {
            synchronized(o.settingsLock) {
                settings = Utils.deepClone(o.settings);
            }
        }
        
        synchronized(statsLock) {
            synchronized(o.statsLock) {
                stats = Utils.deepClone(o.stats);
            }
        }
                
        synchronized(statusLock) {
            synchronized(o.statusLock) {
                status = o.status;
            }
        }
    }
    
    public void doSpeedChange() {
        String method = (String)getSetting("Speed Change Method");
        Object smin = getSetting("Change Min");
        Object smax = getSetting("Change Max");
        Object speedCheck = getSetting("Delay");
        boolean changed = false;
        
                        
        if(rand == null) {
            Object o = getSetting("Change Seed");
            if(o != null) {
                rand = new Random(Utils.makeLong(o));
            }
        }
        
        if(rand != null && speedCheck != null && Utils.makeInt(speedCheck) >= 1) {
            //TODO: Add more null checks
            if(method != null && smin != null && smax != null) {                
                int min = Utils.makeInt(smin);
                int max = Utils.makeInt(smax);
                int change = 0;

                if(min < 1) {
                    min = 1;
                }

                if(method.equals(SettingsConstants.CHANGE_METHOD_CONSTANT)) {
                    change = Utils.makeInt(getSetting("Constant Change Amount"));
                } else if(method.equals(SettingsConstants.CHANGE_METHOD_NORMAL)) {
                    int avg = Utils.makeInt(getSetting("Normal Change Mean"));
                    int stdv = Utils.makeInt(getSetting("Normal Change STDV"));
                    change = (int)Math.round(avg + rand.nextGaussian() * Math.pow(stdv,2));
                } else if(method.equals(SettingsConstants.CHANGE_METHOD_UNIFORM)) {
                    int umin = Utils.makeInt(getSetting("Uniform Change Min"));
                    int umax = Utils.makeInt(getSetting("Uniform Change Max"));
                    change = rand.nextInt((umax - umin) + 1) + umin;
                }

                if(!method.equals(SettingsConstants.CHANGE_METHOD_NONE)) {
                    synchronized(settingsLock) {
                        int speed = Utils.makeInt(settings.get("Delay"));

                        if(change + speed < min) {
                            speed = min;
                            change = 0;
                        }else if(change + speed > max) {
                            speed = max;
                            change = 0;
                        }

                        Log.debug("Tick speed updated to " + (speed + change), toLogID());
                        settings.put("Delay", speed + change);
                        changed = true;
                    }
                }
            }
        }
        
        if(changed) {
            notifyListeners(new Object[]{"Delay", Utils.makeInt(speedCheck)}, NetworkObject.EVENT_SETTINGS_UPDATE, this);
        }
    }
    
    public void destory() {
        synchronized(listeners) {
            listeners.clear();
        }
        
        synchronized(statsLock) {
            stats.clear();
            stats = null;
        }
        
        synchronized(settingsLock) {
            settings.clear();
            settings = null;
        }
    }
    
    protected void resetForLoad() {
        statsLock = new Object();
        settingsLock = new Object();
        status = STATUS_NOTSTARTED;
        statusLock = new Object();
        idLock = new Object();
        listeners = Collections.synchronizedMap(new IdentityHashMap<>());
        rand = null; 
    }
    
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        resetForLoad();
    }
}
