package dans.network;

import dans.util.AlgorithmLoader;
import dans.util.Log;
import dans.util.Utils;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.SwingUtilities;

/**
 *
 * @author Dan
 */
public class NetworkState implements Serializable {
    public static final int EVENT_PROC_ADD = 1;
    public static final int EVENT_PROC_REMOVE = 2;
    public static final int EVENT_PROC_UPDATE_ID = 3;
    
    public static final int EVENT_LINK_ADD = 4;
    public static final int EVENT_LINK_REMOVE = 5;
    public static final int EVENT_LINK_UPDATE = 6;
    
    public static final int EVENT_TICK_UPDATE = 7;
    
    public static final int EVENT_SETTINGS_ADD = 8;
    public static final int EVENT_SETTINGS_UPDATE = 9;
    
    public static final int EVENT_STAT_ADD = 10;
    public static final int EVENT_STAT_UPDATE = 11;
    
    private static final int MIN_TICK_SPEED = 500;
    

    private ConcurrentHashMap<String, Processor> processors = new ConcurrentHashMap<>();
    private ConcurrentHashMap<String, Link> links = new ConcurrentHashMap<>();
    private ConcurrentHashMap<String, Object> networkSettings = new ConcurrentHashMap<>();
    private ConcurrentHashMap<String, Object> networkStats = new ConcurrentHashMap<>();
    
    transient private int tick = 0;
    transient private Object tickLock = new Object();
    
    transient private Map<NetworkStateListener, Boolean> listeners = Collections.synchronizedMap(new IdentityHashMap<>());
    
    transient private ConcurrentHashMap<String, Class> loadedAlgorithms = new ConcurrentHashMap<>();
    private ConcurrentHashMap<String, String> algorithmLocations = new ConcurrentHashMap<>();
    
    transient private Object commonLock = new Object();
    
    transient private ArrayList<String> algorithmLoadFails = null;
    
    private String defaultAlgorthim = null;
    transient private Object defaultAlgorthimLock = new Object();
    
    private NetworkState(){};
    
    public static NetworkState createDefaultState() {
        Log.debug("Creating default network state object...");
        
        NetworkState ns = new NetworkState();

        ns.networkSettings.put("tickSpeed", MIN_TICK_SPEED); //Tick speed in ms
        ns.networkSettings.put("defaultProcSpeed", 1); //Default processor speed in ticks
        ns.networkSettings.put("defaultLinkSpeed", 1); //Default link speed in ticks
        ns.networkSettings.put("defaultProcSpeedChangeMethod", SettingsConstants.CHANGE_METHOD_NONE); //Default method for applying random change in processor speed.
        ns.networkSettings.put("defaultSpeedChangeSeed", ((new Date()).getTime()/1000));
        ns.networkSettings.put("defaultProcSpeedChangeNormMean", 0); //Default mean for noraml distribution change method
        ns.networkSettings.put("defaultProcSpeedChangeNormSTDV", 0); //Default standard deviation for noraml distribution change method
        ns.networkSettings.put("defaultProcSpeedChangeUniMin", 0); //Default minium value for uniform method
        ns.networkSettings.put("defaultProcSpeedChangeUniMax", 0); //Default maxium value for uniform method
        ns.networkSettings.put("defaultProcSpeedChangeMin", 1); //Default minium processor speed in ticks
        ns.networkSettings.put("defaultProcSpeedChangeMax", 30); //Default maxium processor speed in ticks
        ns.networkSettings.put("defaultProcSpeedChangeConstant", 0);
        ns.networkSettings.put("defaultLinkSpeedChangeMethod", SettingsConstants.CHANGE_METHOD_NONE); //Default method for applying random change in link speed.
        ns.networkSettings.put("defaultLinkSpeedChangeNormMean", 0); //Default mean for noraml distribution change method
        ns.networkSettings.put("defaultLinkSpeedChangeNormSTDV", 0); //Default standard deviation for noraml distribution change method
        ns.networkSettings.put("defaultLinkSpeedChangeUniMin", 0); //Default minium value for uniform method
        ns.networkSettings.put("defaultLinkSpeedChangeUniMax", 0); //Default maxium value for uniform method
        ns.networkSettings.put("defaultLinkSpeedChangeMin", 1); //Default minium link speed in ticks
        ns.networkSettings.put("defaultLinkSpeedChangeMax", 30); //Default maxium link speed in ticks
        ns.networkSettings.put("defaultLinkSpeedChangeConstant", 0);
        ns.networkSettings.put("defaultLinkBandwidth", 0.0d); //Default link bandwith (messages per tick), 0 for unlimited.
        ns.networkSettings.put("defaultProcCleanFailRate", 0.0d); //Default processor clean failure rate in percent
        ns.networkSettings.put("defaultProcByzantineFailRate", 0.0d);//Default processor byzantine failure rate in percent
        ns.networkSettings.put("defaultLinkFailRate", 0.0d); //Default link failure rate in percent
        ns.networkSettings.put("defaultLinkLossRate", 0.0d); //Default link packet loss rate in percent
        ns.networkSettings.put("defaultLinkErrorRate", 0.0d); //Default link error rate in percent
        ns.networkSettings.put("defaultAlgorithm", ""); //Name of the default algorithm processors will run
        ns.networkSettings.put("defaultPort", 0); //Default port number for the default algorthim
        ns.networkSettings.put("GUILayout", SettingsConstants.GUI_LAYOUT_NONE); //JGraphX layout to use in GUI.
        ns.networkSettings.put("networkName", "New Network"); //Name of network displayed in the title bar.
        ns.networkSettings.put("loggingLevel", Log.LOG_LEVELS[Log.LOG_LEVEL_DEBUG]); //Logging verbosity.
                
        
        ns.tick = 0;
        
        ns.networkStats.put("Messages Sent", 0); //Number of messages sent
        ns.networkStats.put("Bytes Sent", 0); //Number of bytes sent
        ns.networkStats.put("Cycles", 0); //Number of times a processor has ran its main loop
        ns.networkStats.put("Clean Fails", 0); //Number of processor clean failures
        ns.networkStats.put("Byzantine Fails", 0); //Number of processor byzantine failures
        ns.networkStats.put("Packets Lost", 0); //Number of packets lost
        ns.networkStats.put("Link Fails", 0); //Number of link failures
        ns.networkStats.put("Link Errors", 0); //Number of link errors
        ns.networkStats.put("Active Processors", 0); //Number of active processors
        ns.networkStats.put("Inactive Processors", 0); //Number of inactive processors
        ns.networkStats.put("Terminated Processors", 0); //Number of terminated processors
        ns.networkStats.put("Active Links", 0); //Number of active/working links
        ns.networkStats.put("Inactive Links", 0); //Number of inactive/failed links
        ns.networkStats.put("Network State", SettingsConstants.STATE_STOPPED); //Current simulation state
        
        Log.debug("Default network state object created.");
        
        return ns;
    }
    
    public static NetworkState loadState(String filename) {
        NetworkState ns = new NetworkState();
        
        //TODO: Load file
        
        
        
        return ns;
    }
    
    public synchronized void saveState(String filename) {
        
        //TODO: Save
    }
    
    public Object getCommonLock() {
        return commonLock;
    }
    
    public void addProcessor(Processor p) {
        addProcessor(p, null);
    }
    
    public void addProcessor(Processor p, Object caller) {
        boolean overwrite = false;
        
        synchronized(processors) {
            if(processors.containsKey(p.getID())) {
                overwrite = true;
            }
            processors.put(p.getID(), p);
        }
        
        if(overwrite) {
            Log.debug("WARNING: Processor state overwritten.", "P:" + p.getID());
        }
        
        Log.info("Processor " + p.getID() + " added to network.", "P:" + p.getID());
        notifyProcessorChange(p, NetworkState.EVENT_PROC_ADD, caller);
    }
    
    public void removeProcessor(Processor p) {
        removeProcessor(p.getID(), null);
    }
    
    public void removeProcessor(Processor p, Object caller) {
        removeProcessor(p.getID(), caller);
    }
    
    public void removeProcessor(String id) {
        removeProcessor(id, null);
    }
    
    public void removeProcessor(String id, Object caller) {
        Processor p;
        synchronized(processors) {
            p = processors.remove(id);
        }
        
        if(p == null) {
            Log.debug("Processor " + id + "is not in network state. Could not remove.");
        } else {
            Log.info("Processor " + p.getID() + " removed from network.", "P:" + p.getID());
            notifyProcessorChange(p, NetworkState.EVENT_PROC_REMOVE, caller);
        }
    }
    
    public boolean processorIDUsed(String id) {
        synchronized(processors) {
            return processors.containsKey(id);
        }
    }
    
    public boolean hasProcessor(Processor p) {
        synchronized(processors) {
            return processors.contains(p);
        }
    }
    
    public boolean hasLink(Link l) {
        synchronized(links) {
            return links.contains(l);
        }
    }
    
    public Processor getProcessorByID(String id) {
        synchronized(processors) {
            return processors.get(id);
        }
    }
    
    public boolean updateProcessorID(Processor p, String oldID) {
        return updateProcessorID(p, oldID, null);
    }
    
    public boolean updateProcessorID(Processor p, String oldID, Object caller) {
        boolean changed = false;
        boolean missing = false;
        
        synchronized(processors) {
            if(!processors.containsKey(p.getID())) {       
                if(processors.remove(oldID) == null) {
                    missing = true;
                }
                processors.put(p.getID(), p);
                changed = true;
            }
        }
        
        if(changed && !missing) {
            Log.debug("Processor " + p.getID() + " changed its ID.", "P:" + p.getID());
            notifyProcessorChange(p, NetworkState.EVENT_PROC_UPDATE_ID, caller);
        } else if (changed && missing) {
            Log.info("Processor " + p.getID() + " added to network.", "P:" + p.getID());
            notifyProcessorChange(p, NetworkState.EVENT_PROC_ADD, caller);
        }
        
        return changed;
    }
    
    public Processor[] getProcessors() {
        Processor procs[];
        synchronized(processors) {
            procs = new Processor[processors.size()];
            return processors.values().toArray(procs);
        }
    }
    
    public Link[] getLinks() {
        Link ls[];
        synchronized(links) {
            ls = new Link[links.size()];
            return links.values().toArray(ls);
        }
    }
    
    
    public void addLink(Link l) {
        addLink(l, null);
    }
    
    public void addLink(Link l, Object caller) {
        boolean overwrite = false;
        
        synchronized(links) {
            if(links.containsKey(l.getID())) {
                overwrite = true;
            }
            links.put(l.getID(), l);
        }
        
        if(overwrite) {
            Log.debug("WARNING: Link state overwritten for id " + l.getID());
        }

        Log.info("Link " + l.getID() + " added to network.", "L:" + l.getID());
        notifyLinkChange(l, NetworkState.EVENT_LINK_ADD, caller);
    }
    
    public void removeLink(Link l, Object caller) {
        removeLink(l.getID(), caller);
    }
    
    public void removeLink(Link l) {
        removeLink(l.getID(), null);
    }
    
    public void removeLink(String id, Object caller) {
        Link l;
        synchronized(links) {
            l = links.remove(id);
        }
        
        if(l == null) {
            Log.debug("Link " + id + "is not in network state. Could not remove.");
        } else {
            Log.info("Link " + l.getID() + " removed from network.", "L:" + l.getID());
            notifyLinkChange(l, NetworkState.EVENT_LINK_REMOVE, caller);
        }
    }
    
    public boolean linkIDUsed(String id) {
        synchronized(links) {
            return links.containsKey(id);
        }
    }
    
    public Link getLinkByID(String id) {
        synchronized(links) {
            return links.get(id);
        }
    }
    
     public boolean updateLinkID(Link l, String oldID) {
         return updateLinkID(l, oldID, null);
     }
    
    public boolean updateLinkID(Link l, String oldID, Object caller) {
        boolean changed = false;
        boolean missing = false;
        
        synchronized(links) {
            if(!links.containsKey(l.getID())) {
                if(links.remove(oldID) == null) {
                    missing = true;
                }
                links.put(l.getID(), l);
                changed = true;
            }
        }
        
        if(changed && !missing) {
            Log.debug("Link " + l.getID() + " changed its ID.", "L:" + l.getID());
            notifyLinkChange(l, NetworkState.EVENT_LINK_UPDATE, caller);
        } else if(changed && missing) {
            Log.debug("Link " + l.getID() + " added to network.", "L:" + l.getID());
            notifyLinkChange(l, NetworkState.EVENT_LINK_ADD, caller);
        }
        
        return changed;
    }
       
    public void removeAllLinksOfProcessor(Processor p) {
        removeAllLinksOfProcessor(p, null);
    }
    
    public void removeAllLinksOfProcessor(Processor p, Object caller) {
        Link plinks[] = p.getLinks(this);
        synchronized(links) {
            for(Link l: plinks) {
                links.remove(l.getID());
                Log.info("Link " + p.getID() + " removed from network.", "L:" + p.getID());
                notifyLinkChange(l, NetworkState.EVENT_LINK_REMOVE, caller);
            }
        }
    }
    
    
    public int getTick() {
        synchronized(tickLock) {
            return tick;
        }
    }
    
    public void setTick(int tick) {
        setTick(tick, null);
    }
    
    public void setTick(int tick, Object caller) {
        synchronized(tickLock) {
            this.tick = tick;
        }
        notifyTickChange(tick, NetworkState.EVENT_TICK_UPDATE, caller);
    }
    
    public void incrementTick() {
        incrementTick(null);
    }
    
    public void incrementTick(Object caller) {
        synchronized(tickLock) {
            tick++;
            Log.debug("Tick incremented to " + tick);
        }

        notifyTickChange(tick, NetworkState.EVENT_TICK_UPDATE, caller);
    }
    
    public void addNetworkSetting(String name, Object value) {
        addNetworkSetting(name, value, null);
    }
    
    public void addNetworkSetting(String name, Object value, Object caller) {
        if(!networkSettings.containsKey(name)) {
            synchronized(networkSettings) {
                if(name.equals("tickSpeed") && Utils.makeInt(value) < MIN_TICK_SPEED) {
                    networkSettings.put(name, MIN_TICK_SPEED);
                } else {
                    networkSettings.put(name, value);
                }
            }
            notifySettingsChange(name, value, NetworkState.EVENT_SETTINGS_ADD, caller);
        } else {
            updateNetworkSetting(name, value, caller);
        }
    }
    
    public Object getNetworkSetting(String name) {
        synchronized(networkSettings) {
            return networkSettings.get(name);
        }
    }
    
    public void updateNetworkSetting(String name, Object newvalue) {
        updateNetworkSetting(name, newvalue, null);
    }
    
    public void updateNetworkSetting(String name, Object newvalue, Object caller) {
        synchronized(networkSettings) {
            if(name.equals("tickSpeed") && Utils.makeInt(newvalue) < MIN_TICK_SPEED) {
                networkSettings.put(name, MIN_TICK_SPEED);
            } else {
                networkSettings.put(name, newvalue);
            }
        }
        notifySettingsChange(name, newvalue, NetworkState.EVENT_SETTINGS_UPDATE, caller);
    }
    
    public Object[][] dumpSettings(String keyOrder[]) {
        Object data[][];
        
        synchronized(networkSettings) {
            data = new Object[keyOrder.length][2];
            int i = 0;
            
            for(String name: keyOrder) {
                data[i][0] = name;
                data[i][1] = networkSettings.get(name);
                i++;
            }
        }
        
        return data;
    }
    
    public Object[][] dumpStats(String keyOrder[]) {
        Object data[][];
        
        synchronized(networkStats) {
            data = new Object[keyOrder.length][2];
            int i = 0;
            
            for(String name: keyOrder) {
                data[i][0] = name;
                if(name.equals("Tick")) {
                    data[i][1] = getTick();
                } else {
                    data[i][1] = networkStats.get(name);
                }
                i++;
            }
        }
        
        return data;
    }
    
    public void addNetworkStat(String name, Object value) {
        addNetworkStat(name, value, null);
    }
    
    public void addNetworkStat(String name, Object value, Object caller) {
        boolean changed = false;
        synchronized(networkStats) {
            if(!networkStats.containsKey(name)) {
                networkStats.put(name, value);
            }
        }
        
        if(changed) {
            notifyStatsChange(name, value, NetworkState.EVENT_STAT_ADD, caller);
        }
    }
    
    public Object getNetworkStat(String name) {
        synchronized(networkStats) {
            return networkStats.get(name);
        }
    }
    
    public void updateNetworkStat(String name, Object newvalue) {
        updateNetworkStat(name, newvalue, null);
    }
    
    public void updateNetworkStat(String name, Object newvalue, Object caller) {
        synchronized(networkStats) {
            networkStats.put(name, newvalue);
        }
        notifyStatsChange(name, newvalue, NetworkState.EVENT_STAT_UPDATE, caller);
    }
    
    public void incrementNetworkStat(String name) {
        incrementNetworkStat(name, 1, null);
    }
    
    public void incrementNetworkStat(String name, Object caller) {
        incrementNetworkStat(name, 1, caller);
    }
    
    public void incrementNetworkStat(String name, int by) {
        incrementNetworkStat(name, by, null);
    }
    
    public void incrementNetworkStat(String name, int by, Object caller) {
        boolean changed = false;
        int value = -1;
        
        synchronized(networkStats) {
            Object stat = networkStats.get(name);
            boolean ok = false;
            if(stat != null) {
                try {
                    if(stat instanceof String) {
                        value = Integer.parseInt((String)stat);
                    } else {
                        value = (int)stat;
                    }
                    ok = true;
                } catch(Exception e) {}
                
                if(ok) {
                    networkStats.put(name, (value + by));
                    changed = true;
                }
            }
        }
        
        if(changed) {
            notifyStatsChange(name, (value+by) , NetworkState.EVENT_STAT_UPDATE, caller);
        }
    }
    
    
    public void addListener(NetworkStateListener listener) {
        addListener(listener, false);
    }
    
    public void addListener(NetworkStateListener listener, boolean useInvokeLater) {
        synchronized(listeners) {
            listeners.put(listener, useInvokeLater);
        }
    }
    
    public void removeListener(NetworkStateListener listener) {
        synchronized(listeners) {
            listeners.remove(listener);
        }
    }
    
    private void notifyProcessorChange(Processor p, int eventID, Object caller) {
        synchronized(listeners) {   
            listeners.keySet().stream().forEach((listener) -> {
                if(listener != caller) {
                    if(listeners.get(listener)) {
                        SwingUtilities.invokeLater(() -> {listener.processorChange(p, eventID, caller);});
                    } else {
                        listener.processorChange(p, eventID, caller);
                    }
                }
            });
        }
    }
    private void notifyLinkChange(Link p, int eventID, Object caller) {
        synchronized(listeners) {            
            listeners.keySet().stream().forEach((listener) -> {
                if(listener != caller) {
                    if(listeners.get(listener)) {
                        SwingUtilities.invokeLater(() -> {listener.linkChange(p, eventID, caller);});
                    } else {
                        listener.linkChange(p, eventID, caller);
                    }
                }
            });
        }
    }
    
    private void notifySettingsChange(String name, Object value, int eventID, Object caller) {
        synchronized(listeners) {            
            listeners.keySet().stream().forEach((listener) -> {
                if(listener != caller) {
                    if(listeners.get(listener)) {
                        SwingUtilities.invokeLater(() -> {listener.settingsChange(name, value, eventID, caller);});
                    } else {
                        listener.settingsChange(name, value, eventID, caller);
                    }
                }
            });
        }
    }
    
    private void notifyStatsChange(String name, Object value, int eventID, Object caller) {
        synchronized(listeners) {            
            listeners.keySet().stream().forEach((listener) -> {
                if(listener != caller) {
                    if(listeners.get(listener)) {
                        SwingUtilities.invokeLater(() -> {listener.statsChange(name, value, eventID, caller);});
                    } else {
                        listener.statsChange(name, value, eventID, caller);
                    }
                }
            });
        }
    }
    
    private void notifyTickChange(int tickValue, int eventID, Object caller) {
        synchronized(listeners) {            
            listeners.keySet().stream().forEach((listener) -> {
                if(listener != caller) {
                    if(listeners.get(listener)) {
                        SwingUtilities.invokeLater(() -> {listener.tickChange(tickValue, eventID, caller);});
                    } else {
                        listener.tickChange(tickValue, eventID, caller);
                    }
                }
            });
        }
    }

    public void addListenerToAllProcessors(NetworkManager listener, boolean invokeLater) {
        synchronized(processors) {
            processors.keySet().stream().map((id) -> processors.get(id)).forEach((p) -> {
                p.addListener(listener, invokeLater);
            });
        }
    }

    public void addListenerToAllLinks(NetworkManager listener, boolean invokeLater) {
        synchronized(links) {
            links.keySet().stream().map((id) -> links.get(id)).forEach((l) -> {
                l.addListener(listener, invokeLater);
            });
        }
    }
    
    public void removeListenerToAllProcessors(NetworkManager listener) {
        synchronized(processors) {
            processors.keySet().stream().map((id) -> processors.get(id)).forEach((p) -> {
                p.removeListener(listener);
            });
        }
    }

    public void removeListenerToAllLinks(NetworkManager listener) {
        synchronized(links) {
            links.keySet().stream().map((id) -> links.get(id)).forEach((l) -> {
                l.removeListener(listener);
            });
        }
    }

    public boolean addAlgorithm(Class algorithm, File location) {
        synchronized(loadedAlgorithms) {
            if(!loadedAlgorithms.containsKey(algorithm.getSimpleName())) {
                loadedAlgorithms.put(algorithm.getSimpleName(), algorithm);
                algorithmLocations.put(algorithm.getSimpleName(), location.toString());
                return true;
            }
        }
        
        return false;
    }
    
    public void removeAlgorithm(Class algorithm) {
        removeAlgorithm(algorithm.getSimpleName());
    }
    
    public void removeAlgorithm(String algorithmName) {
        synchronized(loadedAlgorithms) {
            loadedAlgorithms.remove(algorithmName);
        }
    }

    public Class getAlgorithm(String algorithmName) {
        synchronized(loadedAlgorithms) {
            return loadedAlgorithms.get(algorithmName);
        }
    }
    
    public void setDefaultAlgorithm(Class algorithm) {
        if(algorithm != null) {
            setDefaultAlgorithm(algorithm.getSimpleName());
        } else {
            setDefaultAlgorithm("");
        }
    }
    
    public void setDefaultAlgorithm(String algorithm) {
        int defaultPort = Utils.makeInt(getNetworkSetting("defaultPort"));

        synchronized(loadedAlgorithms) {
            if(!algorithm.equals("") && !loadedAlgorithms.containsKey(algorithm)) {
                Log.error("Attempted to set default algorithm to " + algorithm + " which is not in the set of loaded algorithms.");
                return;
            }
        }
        
        updateNetworkSetting("defaultAlgorithm", algorithm, this);
        if(!algorithm.equals("")) {
            Log.info("Default algorithm set to " + algorithm);
        } else {
            Log.info("Default algorithm set to <None>.");
        }
        
        synchronized(processors) {
            processors.values().stream().forEach((p) -> {
                if(!algorithm.equals("")) {
                    if(p.hasAlgorithmSet()) {
                        p.updateAlgorithm(defaultPort, algorithm, this);
                    } else {
                        p.addAlgorithm(defaultPort, algorithm, this);
                    }
                } else {
                    p.removeAlgorithm(defaultPort, this);
                }
            });
        }
        
        synchronized(defaultAlgorthimLock) {
            defaultAlgorthim = algorithm;
        }
    }
    
    public boolean defaultAlgorithmSet() {
        return !getNetworkSetting("defaultAlgorithm").equals("");
    }

    public String[] dumpAlgorithms() {
        synchronized(loadedAlgorithms) {
            String data[] = new String[loadedAlgorithms.size()];
            return loadedAlgorithms.keySet().toArray(data);
        }
    }

    public void resetStats() {
        synchronized(tickLock) {
            tick = 0;
        }
        
        synchronized(networkStats) {
            networkStats.put("Messages Sent", 0); //Number of messages sent
            networkStats.put("Bytes Sent", 0); //Number of bytes sent
            networkStats.put("Cycles", 0); //Number of times a processor has ran its main loop
            networkStats.put("Clean Fails", 0); //Number of processor clean failures
            networkStats.put("Byzantine Fails", 0); //Number of processor byzantine failures
            networkStats.put("Packets Lost", 0); //Number of packets lost
            networkStats.put("Link Fails", 0); //Number of link failures
            networkStats.put("Link Errors", 0); //Number of link errors
            networkStats.put("Active Processors", 0); //Number of active processors
            networkStats.put("Inactive Processors", 0); //Number of inactive processors
            networkStats.put("Terminated Processors", 0); //Number of terminated processors
            networkStats.put("Active Links", 0); //Number of active/working links
            networkStats.put("Inactive Links", 0); //Number of inactive/failed links
            networkStats.put("Network State", SettingsConstants.STATE_STOPPED); //Current simulation state
        }
    }
    
    public void resetSettings() {
        synchronized(networkSettings) {
            networkSettings.put("tickSpeed", MIN_TICK_SPEED); //Tick speed in ms
            networkSettings.put("defaultProcSpeed", 1); //Default processor speed in ticks
            networkSettings.put("defaultLinkSpeed", 1); //Default link speed in ticks
            networkSettings.put("defaultProcSpeedChangeMethod", SettingsConstants.CHANGE_METHOD_NONE); //Default method for applying random change in processor speed.
            networkSettings.put("defaultSpeedChangeSeed", ((new Date()).getTime()/1000));
            networkSettings.put("defaultProcSpeedChangeNormMean", 0); //Default mean for noraml distribution change method
            networkSettings.put("defaultProcSpeedChangeNormSTDV", 0); //Default standard deviation for noraml distribution change method
            networkSettings.put("defaultProcSpeedChangeUniMin", 0); //Default minium value for uniform method
            networkSettings.put("defaultProcSpeedChangeUniMax", 0); //Default maxium value for uniform method
            networkSettings.put("defaultProcSpeedChangeMin", 1); //Default minium processor speed in ticks
            networkSettings.put("defaultProcSpeedChangeMax", 30); //Default maxium processor speed in ticks
            networkSettings.put("defaultProcSpeedChangeConstant", 0);
            networkSettings.put("defaultLinkSpeedChangeMethod", SettingsConstants.CHANGE_METHOD_NONE); //Default method for applying random change in link speed.
            networkSettings.put("defaultLinkSpeedChangeNormMean", 0); //Default mean for noraml distribution change method
            networkSettings.put("defaultLinkSpeedChangeNormSTDV", 0); //Default standard deviation for noraml distribution change method
            networkSettings.put("defaultLinkSpeedChangeUniMin", 0); //Default minium value for uniform method
            networkSettings.put("defaultLinkSpeedChangeUniMax", 0); //Default maxium value for uniform method
            networkSettings.put("defaultLinkSpeedChangeMin", 1); //Default minium link speed in ticks
            networkSettings.put("defaultLinkSpeedChangeMax", 30); //Default maxium link speed in ticks
            networkSettings.put("defaultLinkSpeedChangeConstant", 0);
            networkSettings.put("defaultLinkBandwidth", 0.0d); //Default link bandwith (messages per tick), 0 for unlimited.
            networkSettings.put("defaultProcCleanFailRate", 0.0d); //Default processor clean failure rate in percent
            networkSettings.put("defaultProcByzantineFailRate", 0.0d);//Default processor byzantine failure rate in percent
            networkSettings.put("defaultLinkFailRate", 0.0d); //Default link failure rate in percent
            networkSettings.put("defaultLinkLossRate", 0.0d); //Default link packet loss rate in percent
            networkSettings.put("defaultLinkErrorRate", 0.0d); //Default link error rate in percent
            networkSettings.put("defaultAlgorithm", ""); //Name of the default algorithm processors will run
            networkSettings.put("defaultPort", 0); //Default port number for the default algorthim
            networkSettings.put("GUILayout", SettingsConstants.GUI_LAYOUT_NONE); //JGraphX layout to use in GUI.
            networkSettings.put("networkName", "New Network"); //Name of network displayed in the title bar.
            networkSettings.put("loggingLevel", Log.LOG_LEVELS[Log.LOG_LEVEL_DEBUG]); //Logging verbosity.
        }
    }

    public void clearProcessors() {
        synchronized(processors) {
            for(Processor p: processors.values()) {
                p.destory();
            }
            processors.clear();
        }
    }

    public void clearLinks() {
        synchronized(links) {
            for(Link l: links.values()) {
                l.destory();
            }
            links.clear();
        }
    }

    public void clearAlgorithms() {
        synchronized(loadedAlgorithms) {
            loadedAlgorithms.clear();
            algorithmLocations.clear();
        }
    }
    
    public String getDefaultAlgorthim () {
        synchronized(defaultAlgorthimLock){
            return defaultAlgorthim;
        }
    }
    
    public boolean didAlgorthimsLoadFail() {
        if(algorithmLoadFails == null) {
            return false;
        }
        
        return algorithmLoadFails.size() > 0;
    }
    
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        
        tick = 0;
        tickLock = new Object();
        listeners = Collections.synchronizedMap(new IdentityHashMap<>());
        loadedAlgorithms = new ConcurrentHashMap<>();
        commonLock = new Object();
        algorithmLoadFails = new ArrayList<>();
        defaultAlgorthimLock = new Object();
        
        for(String algo: algorithmLocations.keySet()) {
            File aloc = new File(algorithmLocations.get(algo));
            Class a = null;
            if(aloc.exists() && aloc.canRead()) {
                a = AlgorithmLoader.loadAlgorithm(aloc);
            }
            
            if(a == null) {
                aloc = new File(algo + ".class");
                if(aloc.exists() && aloc.canRead()) {
                    a = AlgorithmLoader.loadAlgorithm(aloc);
                }
            }
            
            if(a != null) {
                loadedAlgorithms.put(algo, a);
            } else {
                algorithmLoadFails.add(algo);
            }
        }
        
        String algofials[] = new String[algorithmLoadFails.size()];
        algofials = algorithmLoadFails.toArray(algofials);
        
        for(Processor p: processors.values()) {
            p.removeAlgorithms(algofials);
        }
    }
            
}
