package dans.network;

import dans.algorithm.Algorithm;
import dans.util.Log;
import dans.util.Utils;
import java.awt.Point;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.concurrent.ConcurrentHashMap;

/**
 *
 * @author Dan
 */
public class Processor extends NetworkObject{   
    public final static int EVENT_NEIGHBOUR_ADDED = 200;
    public final static int EVENT_NEIGHBOUR_REMOVE = 201;
    public final static int EVENT_ALGORITHM_ADDED = 202;
    public final static int EVENT_ALGORITHM_REMOVED = 203;
    public final static int EVENT_ALGORITHM_UPDATED = 204;
    public final static int EVENT_SOURCE_REMOVE = 205;
    public final static int EVENT_SOURCE_ADDED = 206;
    
    public static final int EVENT_MESSAGE_SENT = 210;
    public static final int EVENT_MESSAGE_RECEIVED = 211;
    
    public static final int EVENT_DISPLAY = 220;
    
    public static final int EVENT_PART_FAIL = 230;
        
    
    transient private ConcurrentHashMap<Integer, Algorithm> algorithms = new ConcurrentHashMap<>();
    transient private ConcurrentHashMap<Integer, MessageInQueue> queues = new ConcurrentHashMap<>();
    private ConcurrentHashMap<Integer, String> algorithmNames = new ConcurrentHashMap<>();
    private ConcurrentHashMap<String, Processor> neighbours = new ConcurrentHashMap<>();
    private ConcurrentHashMap<String, Processor> sources = new ConcurrentHashMap<>();
    
    transient private ConcurrentHashMap<String, Processor> sharedMemory = new ConcurrentHashMap<>();
    
    transient private Object algorithmsLock = new Object();
    
    transient private int algorithmsRunning = 0;
    transient private Object algorithmsRunningLock = new Object();
    
    transient private boolean partFail = false;
    
    private Point location = new Point();
    
    public Processor(String id, Processor p) {
        super(id);
        copyFrom(p);
    }
    
    @Override
    protected void copyFrom(NetworkObject o) {
        super.copyFrom(o);
        Processor p = (Processor)o;
        
        synchronized(neighbours) {
            neighbours.clear();
        }
        
        synchronized(sources) {
            sources.clear();
        }
        
        synchronized(algorithmsLock) {
            algorithms.clear();
            queues.clear();
            sharedMemory.clear();
        }
        
        synchronized(algorithmsLock) {
            synchronized(p.algorithmsLock) {
                algorithmNames = Utils.deepClone(p.algorithmNames);
            }
        }
    }
    
    public Processor(String id) {
        super(id);
    }
    
    @Override
    public void setDefaultSettings(NetworkState ns) {
        synchronized(settingsLock) {
            settings.put("Delay", ns.getNetworkSetting("defaultProcSpeed"));
            settings.put("Speed Change Method", ns.getNetworkSetting("defaultProcSpeedChangeMethod"));
            settings.put("Change Min", ns.getNetworkSetting("defaultProcSpeedChangeMin"));
            settings.put("Change Max", ns.getNetworkSetting("defaultProcSpeedChangeMax"));
            settings.put("Normal Change Mean", ns.getNetworkSetting("defaultProcSpeedChangeNormMean"));
            settings.put("Normal Change STDV", ns.getNetworkSetting("defaultProcSpeedChangeNormSTDV"));
            settings.put("Uniform Change Min", ns.getNetworkSetting("defaultProcSpeedChangeUniMin"));
            settings.put("Uniform Change Max", ns.getNetworkSetting("defaultProcSpeedChangeUniMax"));
            settings.put("Constant Change Amount", ns.getNetworkSetting("defaultProcSpeedChangeConstant"));
            settings.put("Clean Failure Rate", ns.getNetworkSetting("defaultProcCleanFailRate"));
            settings.put("Byzantine Failure Rate", ns.getNetworkSetting("defaultProcByzantineFailRate"));
            settings.put("Change Seed", ns.getNetworkSetting("defaultSpeedChangeSeed"));
        }
        
        synchronized(algorithmsLock) {
            algorithms.clear();
            algorithmNames.clear();
            if(ns.defaultAlgorithmSet()) {
                algorithmNames.put(Utils.makeInt(ns.getNetworkSetting("defaultPort")), (String)ns.getNetworkSetting("defaultAlgorithm"));
            }
        }
    }
    
    public String[] getSourceLinkIDs() {
        String links[];
        synchronized(sources) {
            links = new String[sources.size()];
            return sources.keySet().toArray(links);
        }
    }
    
    public Processor[] getSources() {
        Processor procs[];
        synchronized(sources){
            procs = new Processor[sources.size()];
            return sources.values().toArray(procs);
        }
    }
    
    public String[] getSourceIDs() {
        Processor n[] = getSources();
        String ids[] = new String[n.length];
        for(int i = 0; i < n.length; i++) {
            ids[i] = n[i].getID();
        }
        return ids;
    }
    
    public String[] getLinkIDs() {
        String links[];
        synchronized(neighbours) {
            links = new String[neighbours.size()];
            return neighbours.keySet().toArray(links);
        }
    }
    
    public Link[] getLinks(NetworkState ns) {
        String ids[] = getLinkIDs();
        Link links[] = new Link[ids.length];
        for(int i = 0; i < ids.length; i++) {
            links[i] = ns.getLinkByID(ids[i]);
        }
        return links;
    }
    
    public Processor[] getNeighbours() {
        Processor procs[];
        synchronized(neighbours){
            procs = new Processor[neighbours.size()];
            return neighbours.values().toArray(procs);
        }
    }
    
    public String[] getNeighbourIDs() {
        Processor n[] = getNeighbours();
        String ids[] = new String[n.length];
        for(int i = 0; i < n.length; i++) {
            ids[i] = n[i].getID();
        }
        return ids;
    }
    
    public boolean hasNeighbour(String procID) {
        synchronized(neighbours){
            for(Processor p : neighbours.values()) {
                if(p.getID().equals(procID)) {
                    return true;
                }
            }
        }
        
        return false;
    }
    
    public int getLinkDelay(NetworkState ns, String linkID) {
        synchronized(neighbours){
            if(!neighbours.containsKey(linkID)) {
                return -1;
            }
        }
        
        Link link = ns.getLinkByID(linkID);
        if(link == null) {
            return -1;
        }
        
        return Utils.makeInt(link.getSetting("Delay"));
    }
    
    public double getLinkBandwidth(NetworkState ns, String linkID) {
        synchronized(neighbours){
            if(!neighbours.containsKey(linkID)) {
                return -1d;
            }
        }
        
        Link link = ns.getLinkByID(linkID);
        if(link == null) {
            return -1d;
        }
        
        return Utils.makeDouble(link.getSetting("Bandwidth"));
    }
    
    public boolean hasLink(String linkID) {
        synchronized(neighbours){
            return neighbours.containsKey(linkID);
        }
    }
    
    public String getNeighbourID(String linkID) {
        synchronized(neighbours){
            if(neighbours.containsKey(linkID)) {
                return neighbours.get(linkID).getID();
            }
        }
        
        return null;
    }
    
    public String getLinkID(String procID) {
        synchronized(neighbours){
            for(String linkID : neighbours.keySet()) {
                Processor p = neighbours.get(linkID);
                if(p.getID().equals(procID)) {
                    return linkID;
                }
            }
        }
        
        return null;
    }
    
    public void addNeighbour(Link l, Processor p) {
        addNeighbour(l, p, null);
    }
    
    public void addNeighbour(Link l, Processor p, Object caller) {
        boolean changed = false;
        
        synchronized(neighbours){
            if(!neighbours.containsKey(l.getID())) {
                changed = true;
                neighbours.put(l.getID(), p);
            }
        }

        if(changed) {
            if(l.getSource() != this) {
                Log.debug("WARNING: Neighbour added to processor with link " + l.getID() + " that does not list the processor as its source.", "P:" + getID());
            }
            
            notifyListeners(null, EVENT_NEIGHBOUR_ADDED, caller);
            Log.debug("Neighbour " + p.getID() + " added on link " + l.getID(), "P:" + getID());
        } else {
            Log.debug("WARNING: Attempted to overwrite neighbour with id " + p.getID(), "P:" + getID());
        }
    }
    
    public void addSource(Link l, Processor p) {
        addSource(l, p, null);
    }
    
    public void addSource(Link l, Processor p, Object caller) {
        boolean changed = false;
        
        synchronized(sources){
            if(!sources.containsKey(l.getID())) {
                changed = true;
                sources.put(l.getID(), p);
            }
        }

        if(changed) {
            if(l.getTarget() != this) {
                Log.debug("WARNING: Source added to processor with link " + l.getID() + " that does not list the processor as its target.", "P:" + getID());
            }
            
            notifyListeners(null, EVENT_SOURCE_ADDED, caller);
            Log.debug("Source " + p.getID() + " added on link " + l.getID(), "P:" + getID());
        } else {
            Log.debug("WARNING: Attempted to overwrite source with id " + p.getID(), "P:" + getID());
        }
    }
    
    public void removeNeighbour(Link l) {
        removeNeighbour(l, null);
    }
    
    public void removeNeighbour(Link l, Object caller) {
        Processor p;
        
        synchronized(neighbours){
            p = neighbours.remove(l.getID());
        }
        
        if(p != null) {
            notifyListeners(null, EVENT_NEIGHBOUR_REMOVE, caller);
            Log.debug("Neighbour " + p.getID() + " removed on link " + l.getID(), "P:" + getID());
        } else {
            Log.debug("WARNING: Attempted to remove nonexisting neighbour for processor on link " + l.getID(), "P:"+getID());
        }
    }
    
    public void removeSource(Link l) {
        removeSource(l, null);
    }
    
    public void removeSource(Link l, Object caller) {
        Processor p;
        
        synchronized(sources){
            p = sources.remove(l.getID());
        }
        
        if(p != null) {
            notifyListeners(null, EVENT_SOURCE_REMOVE, caller);
            Log.debug("Source " + p.getID() + " removed on link " + l.getID(), "P:" + getID());
        } else {
            Log.debug("WARNING: Attempted to remove nonexisting source for processor on link " + l.getID(), "P:"+getID());
        }
    }
    
    
    @Override
    public void setDefaultStats(NetworkState ns) {
        synchronized(statsLock) {
            stats.put("Messages", 0);
            stats.put("Messages in Bytes", 0);
            stats.put("Cycles", 0);
        }
    }

    @Override
    public String getType() {
        return "Processor";
    }

    public boolean hasAlgorithmSet() {
        synchronized(algorithmsLock) {
            return !algorithmNames.isEmpty();
        }
    }

    public void addAlgorithm(int port, String algorithm) {
        addAlgorithm(port, algorithm, null);
    }
    
    public void addAlgorithm(int port, String algorithm, Object caller) {
        boolean changed = false;
        
        synchronized(algorithmsLock) {
            if(!algorithmNames.containsKey(port)) {
                algorithmNames.put(port, algorithm);
                changed = true;
                Log.debug("Added algorithm " + algorithm + " on port " + port, "P:" + getID());
            } else {
                Log.info("Warning: Attempted to add algorithm to in use port.", "P:" + getID());
            }
        }
        
        if(changed) {
            notifyListeners(new Object[]{port, algorithm}, EVENT_ALGORITHM_UPDATED, caller);
        }
    }

    public void updateAlgorithm(int port, String algorithm) {
        updateAlgorithm(port, algorithm, null);
    }
    
    public void updateAlgorithm(int port, String algorithm, Object caller) {
        boolean changed = false;
        String old = null;
        
        synchronized(algorithmsLock) {
            if(algorithmNames.containsKey(port)) {
                old = algorithmNames.remove(port);
                algorithmNames.put(port, algorithm);
                changed = true;
                Log.debug("Replaced algorithm on port " + port + " with " + algorithm, "P:" + getID());
            } else {
                Log.info("Warning: Attempted to update algorithm on nonexistent port " + port, "P:" + getID());
            }
        }
        
        if(changed) {
            notifyListeners(new Object[]{port, old}, EVENT_ALGORITHM_UPDATED, caller);
        }
    }

    public void removeAlgorithm(int port) {
        removeAlgorithm(port, null);
    }
    
    public void removeAlgorithm(int port, Object caller) {
        boolean changed = false;
        String old = null;
        
        synchronized(algorithmsLock) {
            if(algorithmNames.containsKey(port)) {
                old = algorithmNames.get(port);
                algorithmNames.remove(port);
                changed = true;
                Log.debug("Removed algorithm on port " + port, "P:" + getID());
            } else {
                Log.info("Warning: Attempted to remove algorithm on nonexistent port " + port, "P:" + getID());
            }
        }
        
        if(changed) {
            notifyListeners(new Object[]{port, old}, EVENT_ALGORITHM_REMOVED, caller);
        }
    }
    
    public boolean portInUse(int port) {
        synchronized(algorithmsLock) {
            return algorithmNames.containsKey(port);
        }
    }
    
    public Object[][] dumpAlgorithms() {
        Object data[][];
        
        synchronized(algorithmsLock) {
            data = new Object[algorithmNames.size()][2];
            int i = 0;
            for(int key : algorithmNames.keySet()) {
                data[i][0] = key;
                data[i++][1] = algorithmNames.get(key);
            }
        }
        
        return data;
    }
    
    public void updatePort(int oldPort, int newPort) {
        updatePort(oldPort, newPort, null);
    }
    
    public void updatePort(int oldPort, int newPort, Object caller) {
        boolean changed = false;
        String algorithm = null;
        
        synchronized(algorithmsLock) {
            if(algorithmNames.containsKey(oldPort)) {
                if(!algorithmNames.containsKey(newPort)) {
                    algorithm = algorithmNames.remove(oldPort);
                    algorithmNames.put(newPort, algorithm );
                    changed = true;
                    Log.debug("Updated port of algorithm " + algorithm + " from " + oldPort + " to " + newPort, "P:" + getID());
                } else {
                    Log.info("Warning: Attempted to update port " + oldPort + " to port " + newPort + " that is in use.", "P:" + getID());
                }
            } else {
                Log.info("Warning: Attempted to update port " + oldPort + " that is not in use.", "P:" + getID());
            }
        }
        
        if(changed) {
            notifyListeners(new Object[]{oldPort, algorithm}, EVENT_ALGORITHM_UPDATED, caller);
        }
    }
    
    public void receive(Link l, BackEndMessage bm) {
        String status = getStatus();
        int port = bm.port;
        
        if(!status.equals(NetworkObject.STATUS_ERROR) && !status.equals(NetworkObject.STATUS_FAILED) && !status.equals(NetworkObject.STATUS_TERMINATED)) {
            synchronized(algorithmsLock) {
                MessageInQueue q = queues.get(port);
                if(q != null) {
                    BackEndMessage clone = new BackEndMessage(bm);
                    q.transfer(bm);
                    notifyListeners(new Object[]{l,clone}, EVENT_MESSAGE_RECEIVED, l);
                }
            }
        }
    }
    
    public MessageInQueue getQueue(int port) {
        synchronized(algorithmsLock) {
            return queues.get(port);
        }
    }
    
    public boolean initializeAlgorithms (NetworkState ns){
        int loaded = 0;
        
        synchronized(algorithmsLock) {
            algorithms.clear();
            queues.clear();
            sharedMemory.clear();
            
            try {
                for(int port : algorithmNames.keySet()){
                    String name = algorithmNames.get(port);
                    Log.debug("Setting up algorithm " + name + " on port " + port, "P:" + getID());
                    Class c = ns.getAlgorithm(name);
                    Algorithm instance = (Algorithm) c.newInstance();
                    algorithms.put(port, instance);
                    
                    Method m = Algorithm.class.getDeclaredMethod("setUp", Processor.class, NetworkState.class, Integer.class, ConcurrentHashMap.class);
                    m.setAccessible(true);
                    m.invoke(instance, this, ns, port, sharedMemory);
                    
                    queues.put(port, new MessageInQueue(ns));
                    
                    loaded++;
                }
            } catch(Exception e) {
                Log.error("Failed to initialize algorithm.", e);
                algorithms.clear();
                setStatus(NetworkObject.STATUS_ERROR, this);
                return false;
            }
        }
        
        if(loaded <= 0) {
            Log.info("No algorithm found, stopping." , "P:" + getID());
            setStatus(NetworkObject.STATUS_ERROR, this);
        } else {
            synchronized(algorithmsRunningLock) {
                algorithmsRunning = loaded;
            }
        }
        
        return true;
    }

    public boolean isRoot() {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

    public void startAlgorithms() {
        synchronized(algorithmsLock) {
            for(Algorithm a : algorithms.values()) {
                a.start();
            }
        }
        
        setStatus(NetworkObject.STATUS_RUNNING, this);
    }
    
    public Collection<Algorithm> getAlgorithmInstance() {
        synchronized(algorithmsLock) {
            return algorithms.values();
        }
    }

    public boolean send(NetworkState ns, BackEndMessage bm, Object caller) {
        if(!status.equals(NetworkObject.STATUS_ERROR) && !status.equals(NetworkObject.STATUS_FAILED) && !status.equals(NetworkObject.STATUS_TERMINATED)) {
            synchronized(neighbours) {
                if(bm.linkID != null) {
                    Processor p = neighbours.get(bm.linkID);
                    if(p == null) {
                        return false;
                    }
                    bm.toID = p.getID();
                } else if (bm.toID != null) {
                    for(String linkID : neighbours.keySet()) {
                        Processor p = neighbours.get(linkID);
                        
                        if(p != null && p.getID().equals(bm.toID)) {
                            bm.linkID = linkID;
                            break;
                        }
                    }
                }
            }
                
            if(bm.toID != null && bm.linkID != null) {
                Link l = ns.getLinkByID(bm.linkID);
                BackEndMessage clone = new BackEndMessage(bm);
                boolean sent = l.send(bm);
                
                if(sent) {
                    notifyListeners(clone, EVENT_MESSAGE_SENT, caller);
                }
                
                incrementStat("Messages", this);
                incrementStat("Messages in Bytes", bm.size(), this);
                
                return sent;
            }
        }
        
        return false;
    }

    public void algorithmUpdateStatus(Algorithm a, String status) {

    }
    
    public void algorithmTerminated(Algorithm a) {
        boolean done = false;
        
        synchronized(algorithmsRunningLock) {
            algorithmsRunning--;

            if(algorithmsRunning <= 0) {
                done = true;
            }
            
            if(!partFail && (a.getStatus().equals(NetworkObject.STATUS_FAILED) || a.getStatus().equals(NetworkObject.STATUS_ERROR))) {
                partFail = true;
                notifyListeners(a, EVENT_PART_FAIL, this);
            }
        }
        
        if(done) {
            this.setStatus(checkTermStatus(), this);
        }
    }

    public void displayEvent(String text, int pos, Object caller) {
        notifyListeners(new Object[]{text,pos}, EVENT_DISPLAY, caller);
    }
    
    public char getShortType(){
        return 'P';
    }

    public String checkTermStatus() {
        String statusType = NetworkObject.STATUS_TERMINATED;
        
        synchronized(algorithmsLock) {
            for(Algorithm a : algorithms.values()) {
                String aStatus = a.getStatus();
                if(aStatus.equals(NetworkObject.STATUS_ERROR)) {
                    statusType = NetworkObject.STATUS_ERROR;
                    break;
                } else if(aStatus.equals(NetworkObject.STATUS_FAILED)) {
                    statusType = NetworkObject.STATUS_FAILED;
                }
            }
        }
        
        return statusType;
    }

    int getFailType(int port) {
        Algorithm a;
        
        synchronized(algorithmsLock) {
            a = algorithms.get(port);
        }
        
        if(a == null) {
            return -1;
        }
        
        return a.getFailType();
    }

    void failAlgorithm(int port, boolean clean) {
        Algorithm a;
        
        synchronized(algorithmsLock) {
            a = algorithms.get(port);
        }
        
        if(a != null) {
            a.fail(clean);
        }
    }
    
    @Override
    public void destory() {
        synchronized(algorithmsLock) {
            algorithms.clear();
            queues.clear();
            sharedMemory.clear();
        }
        
        synchronized(neighbours) {
            neighbours.clear();
        }
        
        synchronized(sources) {
            sources.clear();
        }
        
        super.destory();
    }

    void removeAlgorithms(String[] algos) {
        synchronized(algorithmNames) {
            for(String a: algos) {
                for(int port: algorithmNames.keySet()) {
                    if(algorithmNames.get(port).equals(a)) {
                        algorithmNames.remove(port);
                        algorithms.remove(port);
                    }
                }
            }
        }
    }
    
    public void setLocation(Point p) {
        synchronized(location) {
            location.setLocation(p);
        }
    }
    
    public Point getLocation() {
        synchronized(location) {
            return (Point) location.clone();
        }
    }
    
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        resetForLoad();
        algorithms = new ConcurrentHashMap<>();
        queues = new ConcurrentHashMap<>();
        sharedMemory = new ConcurrentHashMap<>();
        algorithmsLock = new Object();
        algorithmsRunning = 0;
        algorithmsRunningLock = new Object();
        partFail = false;
    }
}
