package dans.util;

import static dans.util.Log.makeLog;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.SwingUtilities;

/**
 *
 * @author Dan
 */
public class Log {
    private static Log instance = null;
    
    public static final int LOG_LEVEL_NONE = 0;
    public static final int LOG_LEVEL_ERROR = 1;
    public static final int LOG_LEVEL_INFO = 2;
    public static final int LOG_LEVEL_DEBUG = 3;
     public static final String LOG_LEVELS[] = {"NONE", "ERROR", "INFO", "DEBUG"};
    
    private static final String MAIN_NAME = "MAIN";
    private static final String DIV1 = ":   ";
    private static final String DIV2 = " -> ";
    private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
    
    private PrintStream consoleOut = null;
    private PrintWriter fileOut = null;
    //private GUIProxy guiProxy = null;
    private int logLevel = LOG_LEVEL_DEBUG;
    
    private Lock lock = new ReentrantLock();
    
    private StringBuffer logBuffer = new StringBuffer();
    private ConcurrentHashMap<LogListener, Boolean> listeners = new ConcurrentHashMap<>();
    private Timer timer = new Timer();
    
    private Log() {
        consoleOut = System.out;
        
        try { 
            fileOut = new PrintWriter(new FileWriter("DANS_LOG_" + System.nanoTime() + ".txt"));
        } catch (IOException ex) {}
    }
    
    public static Log makeLog() {
        if(Log.instance == null) {
            Log.instance = new Log();
            debug("Logging started at " + Log.LOG_LEVELS[instance.logLevel] + " level.");
        
            Log.instance.timer.scheduleAtFixedRate(new TimerTask() {
                @Override
                public void run() {
                    Log.updateListeners();
                }
            }, 2500, 1000);
        }        
        
        return Log.instance;
    }
    
    public static synchronized void setConsoleOutput(PrintStream out) {
        makeLog();
        instance.consoleOut = out;
    }
    
    public static synchronized void setFileOutput(PrintWriter out) {
        makeLog();
        instance.fileOut = out;
    }
    
    /*public static synchronized void setGUIOutput(GUIProxy proxy) {
        makeLog();
        instance.guiProxy = proxy;
    }*/
    
    public static synchronized void setLogLevel(String level) {
        makeLog();
        instance.logLevel = java.util.Arrays.asList(Log.LOG_LEVELS).indexOf(level);
        forceMsg("Log level set to " + level);
    }
    
    public static synchronized void setLogLevel(int level) {
        makeLog();
        instance.logLevel = level;
        forceMsg("Log level set to " + level);
    }
    
    public static synchronized void printTrace() {
        for (StackTraceElement ste : Thread.currentThread().getStackTrace()) {
            System.out.println(ste);
        }
    }
    
    private static void forceMsg(String msg) {
        addMsg(msg, Integer.MIN_VALUE);
    }
    
    private static synchronized void addMsg(String msg, int level) {
        makeLog();
        if(instance.logLevel >= level) {
            String log = DATE_FORMAT.format(new Date()) + DIV1 + msg;

            if(instance.consoleOut != null) {
                instance.consoleOut.println(log);
                instance.consoleOut.flush();
            }

            if(instance.fileOut != null) {
                instance.fileOut.println(log);
                instance.fileOut.flush();
            }
            
            instance.logBuffer.append(msg + "\n");
                        
           /* if(instance.guiProxy != null) {
                instance.guiProxy.addToLog(log);
            }*/
        }
    }
    
    public static synchronized void updateListeners() {
        if(instance.listeners.size() > 0 && instance.logBuffer.length() > 0) {
            for(LogListener ll: instance.listeners.keySet()) {
                if(instance.listeners.get(ll)) {
                    String s = instance.logBuffer.toString();
                    SwingUtilities.invokeLater(() -> {ll.logUpdate(s);});
                } else {
                    ll.logUpdate(instance.logBuffer.toString());
                }
            }
            
            instance.logBuffer.setLength(0);
        }
    }
    
    public static synchronized void addListener(LogListener ll) {
        makeLog();
        addListener(ll, false);
    }
    
    public static synchronized void addListener(LogListener ll, boolean useInvokeLater) {
        makeLog();
        instance.listeners.put(ll, useInvokeLater);
    }
    
    public static synchronized void removeListener(LogListener ll) {
        makeLog();
        instance.listeners.remove(ll);
    }
    
    public static synchronized void close() {
        makeLog();
        if(instance.consoleOut != null) {
            instance.consoleOut.close();
            instance.consoleOut = null;
        }

        if(instance.fileOut != null) {
            instance.fileOut.close();
            instance.fileOut.close();
        }
    }
    
    public static void info(String msg) {
        info(msg, null);
    }
    
    public static void info(String msg, String from) {
        if(from == null) {
            from = MAIN_NAME;
        }
        String log = from + DIV2 + msg;
        addMsg(log, Log.LOG_LEVEL_INFO);
    }
    
    public static void debug(String msg) {
        debug(msg, null);
    }
    
    public static void debug(String msg, String from) {
        if(from == null) {
            from = MAIN_NAME;
        }
        String log = from + DIV2 + msg;
        addMsg(log, Log.LOG_LEVEL_DEBUG);
    }
    
    public static void error(String msg) {
        error(msg, null);
    }
    
    public static void error(String msg, Exception e) {
        error(msg, e, null);
    }
    
    public static void error(String msg, Exception e, String from) {
        if(from == null) {
            from = MAIN_NAME;
        }
        
        String log = from + DIV2 + msg;
        
        if(e != null) {
            log += "[" + e.toString() + "]";
            if(instance.logLevel >= Log.LOG_LEVEL_DEBUG) {
                log += "\n" + Arrays.toString(e.getStackTrace());
            }
        }
        addMsg(log, Log.LOG_LEVEL_ERROR);
    }
}
